import pandas as pd
import geopandas as gpd
import numpy as np
import math
import mapbox_earcut as earcut
import vedo
import time
import os
import matplotlib.pyplot as plt 

from tqdm.auto import trange, tqdm
from shapely import affinity
from shapely.ops import linemerge
from shapely.geometry import MultiPolygon, MultiLineString, LineString, Point, MultiPoint, box, polygon, Polygon, LinearRing


import warnings
from shapely.errors import ShapelyDeprecationWarning
warnings.filterwarnings("ignore", category=ShapelyDeprecationWarning) 

class Buildings:
                
    def split_poly(poly, size):

        merged_segments = []
        points = []

        boundaries = list(poly)

        # isCorner = [] # stores which joined_points are corners (list of bool)
        # corners = []

        epsilon = 1e-8
        for poly in boundaries:

            lines = poly.exterior

            length = lines.length
            size = min(size, length)
            num_cells = ((length - length%size)/size) # equal sized cells
            size = length / num_cells
            distances = np.arange(0, length, size)

            interpolated_points = list(MultiPoint([lines.interpolate(distance) for distance in distances]))
            points = list(MultiPoint(lines.coords))

            joined_points = []

            for j in range(0, len(points)-1):
                segment = LineString([points[j], points[j+1]])
                joined_points.append(points[j])
                # corners.append(points[j])

                aux = []
                for k in range(0, len(interpolated_points)):
                    pt = interpolated_points[k]
                    if pt.distance(points[j]) > epsilon and pt.distance(points[j+1]) and pt.buffer(epsilon).intersects(segment):
                        joined_points.append(pt)
                    else:
                        aux.append(pt)
                interpolated_points = aux 

            joined_points.append(points[-1])
            # corners.append(points[-1])

            cur_length = 0
            segments = []
            cur_multiline = []
            for j in range(0, len(joined_points)-1):
                segment = LineString([joined_points[j], joined_points[j+1]])

                cur_multiline.append(segment)
                cur_length += segment.length

                if (cur_length >= size-1e-5):
                    cur_length = 0
                    cur_pt = joined_points[j]
                    segments.append(cur_multiline)
                    cur_multiline = []

    #             if (cur_length < DISTANCE_DELTA-1e-5) and (j == len(joined_points)-2):
    #                 segments[-1].append(segment)

            # for j in range(0,len(segments)):
            #     seg = linemerge(MultiLineString(segments[j]))
            #     for coordObj in list(MultiPoint(seg.coords)): # TODO: maybe we can have performance issues in this nested loop
            #         corner = False
            #         for elem in corners:
            #             if(coordObj.equals(elem)):
            #                 corner = True
            #                 break
            #         if corner:
            #             # if(len(isCorner) >= 2):
            #             #     # isCorner[len(isCorner)-2] = False
            #             #     if isCorner[len(isCorner)-2]: # You can not have two trues in a row
            #             #         isCorner.append(False)
            #             #     else:
            #             #         isCorner.append(True)
            #             isCorner.append(True)
            #         else:
            #             isCorner.append(False)

            #     merged_segments.append(seg)

            for j in range(0,len(segments)):
                seg = linemerge(MultiLineString(segments[j]))
                merged_segments.append(seg)

            points = joined_points

        # return merged_segments, isCorner #, list(MultiPoint([lines.interpolate(distance) for distance in distances]))
        return merged_segments #, list(MultiPoint([lines.interpolate(distance) for distance in distances]))
    

    # def extrude(segments, min_height, height, size, isCorner):
    def extrude(segments, min_height, height, size):

        length = height-min_height
        
        if size == -1:
            size = length
        
        size = min(size, length)
        num_cells = ((length - length%size)/size) # equal sized cells

        size = length/num_cells
        distances = np.arange(0, length, size)
#         print(min_height, height, distances)

        coordinates = []
        indices = []
        ids = []
        colors = []
        # corner = [] # one value for each coordinate
        # width = []
        # heights = [] # one value for each coordinate
        footPrintPoints = [] # stores all footprint points that form the segments
        groupedFootPrintPoints = [] # stores all footprint points grouped by segment
        # duplicatedIsCorner = []
        # cornersIndex = [] # stores the index of the corners in the polygon. Two indices define a facade.
        cid = -1

        for index, seg in enumerate(segments):
            
            points = list(zip(*seg.coords.xy))

            footPrintPoints += points

            groupedFootPrintPoints.append(points)

        # # duplicates all corners (if they are not already duplicated)
        # duplicatedPoints = []
        # readSegPoints = 0
        # for segPoints in groupedFootPrintPoints:
        #     duplicatedPoints.append([])
        #     for index, point in enumerate(segPoints):
        #         duplicatedPoints[len(duplicatedPoints)-1].append(point)
        #         duplicatedIsCorner.append(isCorner[readSegPoints+index])
        #         if isCorner[readSegPoints+index]:
        #             if(index != 0 and index != len(segPoints)-1): # if point is in the beggining or end of the segment it is already duplicated
        #                 duplicatedPoints[len(duplicatedPoints)-1].append(point) # TODO maybe copy the point array
        #                 duplicatedIsCorner.append(isCorner[readSegPoints+index])

        #     readSegPoints += len(segPoints)

        # flatDuplicatedPoints = [item for sublist in duplicatedPoints for item in sublist]

        # uvPerPoint = np.empty(len(flatDuplicatedPoints)) # stores the uv value for each point in the segments 
        # wallsWidth = np.empty(len(flatDuplicatedPoints)) # stores the width of all walls determined by two corners

        # def calcDistance(listPoints, point1, point2):
        #     if(point1 == point2):
        #         return 0
            
        #     accDistance = 0
            
        #     if(point1 < point2):
        #         for i in range(point1, point2):
        #             accDistance += math.sqrt(math.pow(listPoints[i+1][0] - listPoints[i][0], 2) + math.pow(listPoints[i+1][1] - listPoints[i][1], 2))
        #     else:
        #         # calculate distance of all points after point1
        #         for i in range(point1, len(listPoints)-1):
        #             accDistance += math.sqrt(math.pow(listPoints[i+1][0] - listPoints[i][0], 2) + math.pow(listPoints[i+1][1] - listPoints[i][1], 2))

        #         # calculate distance of all points before point2
        #         for i in range(0, point2):
        #             accDistance += math.sqrt(math.pow(listPoints[i+1][0] - listPoints[i][0], 2) + math.pow(listPoints[i+1][1] - listPoints[i][1], 2))

        #     return accDistance

        # def getUV(listPoints, anchor1, p):

        #     distanceAnchor1 = calcDistance(listPoints, anchor1, p)
            
        #     return distanceAnchor1

        # cornersPairs = [] # store tuples containing pairs of corners that define a wall
        # cornerIndex = 0
        # firstCorner = -1
        # for index, value in enumerate(duplicatedIsCorner): # len(duplicatedIsCorner) == len(flatDuplicatedPoints)
        #     if value:
        #         if (not duplicatedIsCorner[0]) and cornerIndex == 0: # if the first is not a coordinate skip the first corner for now
        #             firstCorner = index
        #             continue
        #         if cornerIndex%2 == 0:
        #             cornersPairs.append([]) # creates another tuple
        #             cornersPairs[len(cornersPairs)-1].append(index) # the first position is the corner that will receive 0
        #         else:
        #             cornersPairs[len(cornersPairs)-1].append(index) # the second position is the corner that will receive 1
        #         cornerIndex += 1

        # if not duplicatedIsCorner[0]: # if the first point is not a coordinate an extra pair connecting the ending with the beggining needs to be added
        #     cornersPairs[len(cornersPairs)-1].append(firstCorner)

        # maxDistance = -1
        # for pair in cornersPairs:
        #     calcDist = calcDistance(flatDuplicatedPoints, pair[0], pair[1])
        #     if(calcDist > maxDistance):
        #         maxDistance = calcDist

        # for pair in cornersPairs:
        #     for indexPoint in range(pair[0], pair[1]+1):
        #         wallWidth = calcDistance(flatDuplicatedPoints, pair[0], pair[1])
        #         wallsWidth[indexPoint] = wallWidth

        # # for each pair of corners calculates uv for all points between the corners
        # for pair in cornersPairs:
        #     if(pair[1] > pair[0]): # it is a pair in the middle of the sequence
        #         for indexPoint in range(pair[0]+1, pair[1]):
        #             uv = getUV(flatDuplicatedPoints, pair[0], indexPoint)
        #             uvPerPoint[indexPoint] = uv
        #     elif (pair[1] < pair[0]): # it is a pair in the end of the sequence
        #         # calculate uv for every point greater than the last corner
        #         for indexPoint in range(pair[0]+1, len(flatDuplicatedPoints)):
        #             uv = getUV(flatDuplicatedPoints, pair[0], indexPoint)
        #             uvPerPoint[indexPoint] = uv

        #         # calculate uv for every point smaller than the first corner
        #         for indexPoint in range(0, pair[1]):
        #             uv = getUV(flatDuplicatedPoints, pair[0], indexPoint)
        #             uvPerPoint[indexPoint] = uv

        #     uvPerPoint[pair[0]] = 0
        #     uvPerPoint[pair[1]] = calcDistance(flatDuplicatedPoints, pair[0], pair[1])

        alreadySeen = 0
        # for index, seg in enumerate(duplicatedPoints):
        for seg in segments:
            # points = seg
            for distance in distances:
                if distance + min_height >= height:
                    continue
                points = list(zip(*seg.coords.xy))
                color = np.random.rand(3,)
                cid+=1
                for i in range(0, len(points)-1):

                    p0 = points[i]
                    p1 = points[i+1]

                    if Point(p0).equals(Point(p1)): # it is one of the points previously duplicated
                        continue

                    v0 = [p0[0],p0[1],distance+min_height]
                    v1 = [p1[0],p1[1],distance+min_height]
                    v2 = [p0[0],p0[1],distance+min_height+size]
                    v3 = [p1[0],p1[1],distance+min_height+size]

                    i0 = int(len(coordinates)/3)
                    coordinates.extend(v0)
                    coordinates.extend(v1)
                    coordinates.extend(v2)
                    coordinates.extend(v3)

                    # corner.append([uvPerPoint[alreadySeen+i], v0[2]-min_height])
                    # corner.append([uvPerPoint[alreadySeen+i+1], v1[2]-min_height])
                    # corner.append([uvPerPoint[alreadySeen+i], v2[2]-min_height])
                    # corner.append([uvPerPoint[alreadySeen+i+1], v3[2]-min_height])

                    # corner.append(uvPerPoint[alreadySeen+i])
                    # corner.append(uvPerPoint[alreadySeen+i+1])
                    # corner.append(uvPerPoint[alreadySeen+i])
                    # corner.append(uvPerPoint[alreadySeen+i+1])

                    # width.append(wallsWidth[alreadySeen+i])
                    # width.append(wallsWidth[alreadySeen+i+1])
                    # width.append(wallsWidth[alreadySeen+i])
                    # width.append(wallsWidth[alreadySeen+i+1])

                    # heights.append(length)
                    # heights.append(length)
                    # heights.append(length)
                    # heights.append(length)

                    # t0
                    indices.append(i0)
                    indices.append(i0+1)
                    indices.append(i0+2)

                    # t1
                    indices.append(i0+1)
                    indices.append(i0+3)
                    indices.append(i0+2)
                    
                    # color and ids per face
                    colors.append(color)
                    colors.append(color)
                    ids.append(cid)
                    ids.append(cid)
            
            alreadySeen += len(points)

        coordinates = np.array(coordinates).reshape(-1, 3)
        indices = np.array(indices).reshape(-1, 3)
        colors = np.array(colors).reshape(-1, 3)
        ids = np.array(ids)
        # corner = np.array(corner)
        # width = np.array(width)
        # heights = np.array(heights)

        # return coordinates, indices, ids, colors, corner, width, heights
        return coordinates, indices, ids, colors

    def azimuth(mrr):

        def _azimuth(point1, point2):
            """azimuth between 2 points (interval 0 - 180)"""
            angle = np.arctan2(point2[0] - point1[0], point2[1] - point1[1])
            return np.degrees(angle) if angle > 0 else np.degrees(angle) + 180

        def _dist(a, b):
            """distance between points"""
            return math.hypot(b[0] - a[0], b[1] - a[1])

        """azimuth of minimum_rotated_rectangle"""
    #     bbox = list(mrr.exterior.coords)
        bbox = list(mrr.coords)
        axis1 = _dist(bbox[0], bbox[3])
        axis2 = _dist(bbox[0], bbox[1])

        if axis1 <= axis2:
            az = _azimuth(bbox[0], bbox[1])
        else:
            az = _azimuth(bbox[0], bbox[3])

        return az

    def get_roof(bottom_poly, top_poly, height, size):
        
    #     gdf_poly = gpd.GeoDataFrame(geometry=[poly])
        if top_poly != None:
            polys = bottom_poly.difference(top_poly)
        else:
            polys = bottom_poly
            
        if polys.geom_type == 'Polygon':
            polys = MultiPolygon([polys])

        coordinates = []
        indices = []
        ids = []
        colors = []
        # uv = []
        # roofWidth = []
        # roofHeights = []

        cid = -1
        count = 0
        # for poly in polys:   
        for i in range(len(polys)):   

            poly = polys[i]

            mrr = poly.minimum_rotated_rectangle
            
            rot = Buildings.azimuth(mrr.exterior)
            poly = affinity.rotate(poly, rot, (0,0))

            line = LineString(poly.exterior.coords)
            gdf_line = gpd.GeoDataFrame(geometry=[line])

            xmin, ymin, xmax, ymax = gdf_line.total_bounds
            
            if size == -1:
                cells = [box(xmin, ymin, xmax, ymax)]
            else:
                cell_width = cell_height = size
                cells = []
                for x0 in np.arange(xmin, xmax+cell_width, cell_width):
                    for y0 in np.arange(ymin, ymax+cell_height, cell_height):
                        x1 = x0-cell_width
                        y1 = y0+cell_height
                        new_cell = box(x0, y0, x1, y1)
                        cells.append(new_cell)

            intersection = gpd.overlay(gpd.GeoDataFrame({'geometry': gpd.GeoSeries(poly)}), gpd.GeoDataFrame({'geometry': gpd.GeoSeries(cells)}),how='intersection',keep_geom_type=True)

            # # collecting width data while the plane is aligned with x and y
            # for elem in intersection.values:
            #     cell = elem[0]
            #     ccell = []
            #     if cell.geom_type == 'MultiPolygon':
            #         ccell = list(cell)
            #     elif cell.geom_type == 'Polygon':
            #         ccell = [cell]

            #     for c in ccell:
            #         points = np.array(c.exterior.coords[0:-1]) # remove last one (repeated)
            #         points = points.flatten().tolist()

            #         for i in range(0, len(points), 2):
            #             uv.append(points[i+1]-xmin) # append the distance between the i point and xmin
            #             roofWidth.append(0.5) # appned the width of the roof at that point

            cells = intersection.rotate(-rot, origin=(0,0)).values

            # for cell in cells:
            for j in range(len(cells)):

                cell = cells[j]

                ccell = []
                if cell.geom_type == 'MultiPolygon':
                    ccell = list(cell)
                elif cell.geom_type == 'Polygon':
                    ccell = [cell]

                for c in ccell:
        
                    points = np.array(c.exterior.coords[0:-1]) # remove last one (repeated)
                    rings = np.array([len(points)])

                    ind = earcut.triangulate_float64(points, rings)

                    # print(points)
                    # print(ind)
                    ind = (ind+count).tolist()
                    indices += ind

                    color = np.random.rand(3,)
                    points = points.flatten().tolist()
                    cid+=1
                    for i in range(0, len(points), 2):
                        coordinates.append(points[i])
                        coordinates.append(points[i+1])
                        coordinates.append(height)

                        # uv.append(-1.0)
                        # roofWidth.append(-1.0)
                        # roofHeights.append(-1.0)

                    count = int(len(coordinates)/3)
                    
                    colors = colors + [color] * int(len(ind)/3)
                    ids = ids + [cid] * int(len(ind)/3)


        # end = time.time()
        # print("total:"+str(end - start))

        coordinates = np.array(coordinates).reshape(-1, 3)
        indices = np.array(indices).reshape(-1, 3)
        colors = np.array(colors).reshape(-1, 3)
        ids = np.array(ids)
        # uv = np.array(uv).reshape(-1, 2)
        # uv = np.array(uv)
        # roofWidth = np.array(roofWidth)
        # roofHeights = np.array(roofHeights)

        # return coordinates, indices, ids, colors, uv, roofWidth, roofHeights
        return coordinates, indices, ids, colors

    def merge_building_blocks(building):
        '''
            Groups the blocks that compose one building in the OSM elements
        '''

        building = building.to_crs('epsg:3395')

        min_heights = building.min_height.values
        heights = building.height.values
        all_heights = np.concatenate((min_heights, heights))
        all_heights = np.unique(all_heights)
        all_heights = np.sort(all_heights)

        merged_geometry = []
        merged_height = []
        merged_min_height = []
        for i in range(0, len(all_heights)-1):
            min_height = all_heights[i]
            height = all_heights[i+1]
            bbl = building[(building['min_height'] <= all_heights[i]) & (building['height'] > all_heights[i])]

            if len(bbl) == 0:
                continue

            merged_geom = bbl.unary_union
            if merged_geom.type == 'Polygon':
                merged_geom = MultiPolygon([merged_geom])

            oriented_geom = []
            for geom in merged_geom:
                oriented_geom.append(polygon.orient(geom, 1)) # counter-clockwise
            oriented_geom = MultiPolygon(oriented_geom)

            merged_geometry.append(oriented_geom)
            merged_min_height.append(min_height)
            merged_height.append(height)

        d = {'min_height': merged_min_height, 'height': merged_height, 'geometry': merged_geometry}
        merged_building = gpd.GeoDataFrame(d, crs='epsg:3395')
        merged_building['building_id'] = building.iloc[0]['building_id']

        return merged_building

    def create_building_mesh(building, size = -1):

        merged_building = Buildings.merge_building_blocks(building)

        boundaries = list(merged_building.geometry)

        coords = np.empty((0,3))
        indices = np.empty((0,3))
        ids = np.empty((0))
        colors = np.empty((0,3))
        # uv = np.empty((0))
        # width = np.empty((0))
        # heights = np.empty((0))
        # pointsPerSection = []

        orientedEnvelope = []
        sectionFootprint = []
        # sectionHeight = []
        # sectionMinHeight = []

        for i in range(0, len(boundaries)):

            geom = boundaries[i].geoms
            height = merged_building.iloc[i].height
            min_height = merged_building.iloc[i].min_height

            # sectionHeight.append(height)
            # sectionMinHeight.append(min_height)

            orientedEnvelope.append([])
            sectionFootprint.append([])

            for index, coordinate in enumerate(geom[0].minimum_rotated_rectangle.exterior.coords): # TODO: check if only one geometry is stored in geom
                if(index != (len(geom[0].minimum_rotated_rectangle.exterior.coords) - 1)):
                    orientedEnvelope[len(orientedEnvelope)-1] += list(coordinate)

            for coordinate in geom[0].exterior.coords:
                sectionFootprint[len(sectionFootprint)-1] += list(coordinate)

            # roof
            bottom_poly = boundaries[i]
            if i + 1 < len(boundaries):
                top_poly = boundaries[i+1]
            else:
                top_poly = None

            # coordinates_roof, indices_roof, ids_roof, colors_roof, uv_roof, roof_width, roof_heights = Buildings.get_roof(bottom_poly, top_poly, height, size)
            coordinates_roof, indices_roof, ids_roof, colors_roof = Buildings.get_roof(bottom_poly, top_poly, height, size)
            coordinates_roof = coordinates_roof.reshape(-1, 3)

            indices_roof = indices_roof.reshape(-1, 3)
            indices_roof = indices_roof + coords.shape[0]
            ids_roof = ids_roof + ids.shape[0]
            
            # walls
            # segments, isCorner = Buildings.split_poly(geom, size)
            segments = Buildings.split_poly(geom, size)

            # coordinates_walls, indices_walls, ids_walls, colors_walls, uv_walls, walls_width, walls_heights = Buildings.extrude(segments, min_height, height, size, isCorner)
            coordinates_walls, indices_walls, ids_walls, colors_walls = Buildings.extrude(segments, min_height, height, size)
            
            indices_walls = indices_walls + coordinates_roof.shape[0] + coords.shape[0]
            ids_walls = ids_walls + ids_roof.shape[0] + ids.shape[0]

            coordinates_walls = coordinates_walls.reshape(-1, 3)
            indices_walls = indices_walls.reshape(-1, 3)

            coords = np.concatenate((coords, coordinates_roof, coordinates_walls))
            indices = np.concatenate((indices, indices_roof, indices_walls))
            ids = np.concatenate((ids, ids_roof, ids_walls))
            colors = np.concatenate((colors, colors_roof, colors_walls))

            # uv = np.concatenate((uv, uv_roof, uv_walls))

            # width = np.concatenate((width, roof_width, walls_width))

            # heights = np.concatenate((heights, roof_heights, walls_heights))

            # pointsPerSection.append(len(coordinates_roof)+len(coordinates_walls))


        coords = coords.reshape(-1, 3)
        indices = indices.reshape(-1, 3)

        # pointsPerSection = np.array(pointsPerSection)

        # return coords, indices, ids, colors, sectionHeight, sectionMinHeight, orientedEnvelope, sectionFootprint, uv, width, heights, pointsPerSection
        return coords, indices, ids, colors, orientedEnvelope, sectionFootprint

    def generate_building_layer(gdf, size):
        
        gdf = gdf.to_crs('epsg:3395')   

        building_ids = []
        coordinates = []
        indices = []
        ids = []
        colors = []
        # heights = []
        # minHeights = []
        envelopes = []
        footprints = []
        # allCorners = []
        # allWidth = []
        # surfaceHeights = []
        # pointsPerSection = []
        unique_buildings = gdf.index.unique()

        # for i in trange(len(unique_buildings)):
        for i in range(len(unique_buildings)):
            building_id = unique_buildings[i]
            building = gdf.loc[[building_id]]
            # coord, ind, iids, cols, sectionHeight, sectionMinHeight, orientedEnvelope, sectionFootprint, corners, width, surfaceHeight, pointsSection = Buildings.create_building_mesh(building, size)
            coord, ind, iids, cols, orientedEnvelope, sectionFootprint = Buildings.create_building_mesh(building, size)
            building_ids.append(building_id)
            coordinates.append(coord)
            indices.append(ind)
            ids.append(iids)
            colors.append(cols)
            # heights.append(sectionHeight)
            # minHeights.append(sectionMinHeight)
            envelopes.append(orientedEnvelope)
            footprints.append(sectionFootprint)
            # allCorners.append(corners)
            # allWidth.append(width)
            # surfaceHeights.append(surfaceHeight)
            # pointsPerSection.append(pointsSection)

        # df = pd.DataFrame({
        #     'building_id': pd.Series(building_ids),
        #     'coordinates': pd.Series(coordinates),
        #     'indices': pd.Series(indices),
        #     'ids': pd.Series(ids),
        #     'colors': pd.Series(colors),
        #     'heights': pd.Series(heights),
        #     'minHeights': pd.Series(minHeights),
        #     'orientedEnvelope': pd.Series(envelopes),
        #     'sectionFootprint': pd.Series(footprints),
        #     'uv': pd.Series(allCorners),
        #     'width': pd.Series(allWidth),
        #     'surfaceHeight': pd.Series(surfaceHeights),
        #     'pointsPerSection': pd.Series(pointsPerSection)
        # })

        df = pd.DataFrame({
            'building_id': pd.Series(building_ids),
            'coordinates': pd.Series(coordinates),
            'indices': pd.Series(indices),
            'ids': pd.Series(ids),
            'colors': pd.Series(colors),
            'orientedEnvelope': pd.Series(envelopes),
            'sectionFootprint': pd.Series(footprints),
        })

        tridimensional_coordinates = []
        ids_tridimensional_coordinates = []
        counter_id_tridimensional_coordinates = 0

        for sublist in coordinates:
            for elem in sublist:
                tridimensional_coordinates.append(elem)
                ids_tridimensional_coordinates.append(counter_id_tridimensional_coordinates)
                counter_id_tridimensional_coordinates += 1

        geometries = []
        ids = [] 
        geometries_coordinates = []
        ids_coordinates = []
        counter_id_coordinates = 0

        for id, elem in enumerate(footprints):

            ids.append(id)

            polygon_coordinates = elem[0]
            groupedCoordinates = []

            for i in range(0,int(len(polygon_coordinates)/2)):
                groupedCoordinates.append((polygon_coordinates[i*2], polygon_coordinates[i*2+1]))
                geometries_coordinates.append(Point(polygon_coordinates[i*2], polygon_coordinates[i*2+1]))
                ids_coordinates.append(counter_id_coordinates)
                counter_id_coordinates += 1

            geometries.append(Polygon(groupedCoordinates))

        gdf = gpd.GeoDataFrame({'geometry': geometries, 'id': ids}, crs=3395)

        gdf_coordinates = gpd.GeoDataFrame({'geometry': geometries_coordinates, 'id': ids_coordinates}, crs=3395)
        df_3d_coordinates = pd.DataFrame({'geometry': tridimensional_coordinates, 'id': ids_tridimensional_coordinates})

        df = df.set_index('building_id', drop=False)
        # df = df.sort_index()

        return {"df": df, "gdf": {"objects": gdf, "coordinates": gdf_coordinates, "coordinates3d": df_3d_coordinates}}

    def get_coordinates(gdf, compute_normals=False):
        coordinates = gdf['coordinates'].values
        indices = gdf['indices'].values
        ids = gdf['ids'].values
        colors = gdf['colors'].values
        
        coords_all = np.empty((0,3))
        indices_all = np.empty((0,3))
        colors_all = np.empty((0,3))
        ids_all = np.empty((0))
        
        for i in range(0, len(coordinates)):
            indices_all = np.concatenate((indices_all, np.array(indices[i])+coords_all.shape[0]))
            coords_all = np.concatenate((coords_all, coordinates[i]))
            colors_all = np.concatenate((colors_all, colors[i]))
            ids_all = np.concatenate((ids_all, ids[i]))
            
        if compute_normals:
            vmesh = vedo.Mesh([coords_all, indices_all])
            normals_all = vmesh.normals(cells=False)
            return coords_all, indices_all, ids_all, colors_all, normals_all
        
        return coords_all, indices_all, ids_all, colors_all
    
    def df_to_json(df, layer_id = "buildings", layer_type = 'BUILDINGS_LAYER', renderStyle = ["SMOOTH_COLOR"], styleKey = "building"):

        json_new = {}

        # config parameters
        json_new["id"] = layer_id
        json_new["type"] = layer_type
        json_new["renderStyle"] = renderStyle
        json_new["styleKey"] = styleKey

        json_new["data"] = []

        for index in range(0, len(df)):

            flattened_coordinates = []
            flattened_indices = []
            flattened_normals = []

            coords_all, indices_all, ids_all, _, normals = Buildings.get_coordinates(df.iloc[[index]], compute_normals=True) # Calculating normals

            flattened_coordinates += [round(item,4) for sublist in coords_all for item in sublist]
            flattened_indices += [int(item) for sublist in indices_all for item in sublist]
            flattened_normals += [round(float(item),4) for sublist in normals for item in sublist]

            # json_new["data"].append({
            #     "geometry": {
            #         "coordinates": flattened_coordinates,
            #         "indices": flattened_indices,
            #         "normals": flattened_normals,
            #         "ids": [int(elem) for elem in ids_all],
            #         "heights": [round(item,6) for item in df.iloc[[index]]["heights"].tolist()[0]],
            #         "minHeights": [round(item,6) for item in df.iloc[[index]]["minHeights"].tolist()[0]],
            #         "orientedEnvelope": [[round(elem,6) for elem in item] for item in df.iloc[[index]]["orientedEnvelope"].tolist()[0]],
            #         "sectionFootprint": [[round(elem,6) for elem in item] for item in df.iloc[[index]]["sectionFootprint"].tolist()[0]],
            #         "uv": [round(item, 6) for item in df.iloc[[index]]["uv"].tolist()[0].tolist()],
            #         "width": [round(item,6) for item in df.iloc[[index]]["width"].tolist()[0].tolist()],
            #         "pointsPerSection": df.iloc[[index]]["pointsPerSection"].tolist()[0].tolist()
            #     }
            # })

            json_new["data"].append({
                "geometry": {
                    "coordinates": flattened_coordinates,
                    "indices": flattened_indices,
                    "normals": flattened_normals,
                    "ids": [int(elem) for elem in ids_all],
                    "orientedEnvelope": [[round(elem,4) for elem in item] for item in df.iloc[[index]]["orientedEnvelope"].tolist()[0]],
                    "sectionFootprint": [[round(elem,4) for elem in item] for item in df.iloc[[index]]["sectionFootprint"].tolist()[0]]
                }
            })

        return json_new

    def merge_overlapping_buildings(gdf):
        '''
            Groups the buildings that overlap

            Args:
                gdf (GeoDataFrame): A GeoDataFrame with the atributes 'building_id', 'geometry', 'min_height', 'height', 'tags' describing each block that composes one building

            Returns:
                result (GeoDataFrame): A GeoDataFrame with the same atributes but with the overlaping buildings merged
        '''

        gdf = gdf.to_crs('epsg:3395')   

        # merge buildings that overlap
        unique_buildings = gdf.index.unique()
        for i in range(len(unique_buildings)):
            building_id = unique_buildings[i]
            buildings = gdf[gdf['building_id']==building_id]
            if(len(buildings) > 0):
                contained = gdf.sindex.query(buildings.geometry.unary_union, predicate='intersects')
                iid = gdf.iloc[contained]['building_id'].values[0]
                gdf.iloc[contained, gdf.columns.get_loc('building_id')] = iid
        gdf = gdf.set_index('building_id', drop=False)
        gdf = gdf.sort_index()

        gdf = gdf.to_crs('epsg:4326')
        return gdf


import json
import geopandas as gpd
import pandas as pd
import os
import struct
import webbrowser

from shapely.geometry import Polygon, Point

class UrbanComponent:
    """
    Basic Urban Toolkit component
    """

    cid = None
    style = {}
    layers = {'json': [], 'gdf': {'objects': [], 'coordinates': [], 'coordinates3d': []}}
    joinedJson = {}
    camera = None
    bpolygon = []
    workDir = None

    def __init__(self, cid = 'map', filepath = None, layers = None, camera = None, bpolygon = None):
        if filepath != None:
            self.from_file(filepath)
        self.cid = cid
        if layers != None:
            self.layers = layers
        if camera != None:
            self.camera = camera
        if bpolygon != None:
            self.bpolygon = bpolygon

    def set_work_dir(self, dir):
        self.workDir = dir
        
    # TODO: make this function more generic regarding the section footprint
    def jsonToGdf(self, layer_json, dim, abstract=False):

        ids = []
        ids_coordinates = []
        values_coordinates = []
        counter_id_coordinates = 0

        geometries = []
        geometries_coordinates = []

        tridimensional_coordinates = []
        ids_tridimensional_coordinates = []
        counter_id_tridimensional_coordinates = 0

        dimensions = 3


        if(not abstract):
            if('sectionFootprint' in layer_json['data'][0]['geometry']): # hard coded buildings case. We want to consider the footprint for 2D joins not the whole building
                dimensions = 2 

            for id, elem in enumerate(layer_json['data']):

                groupedCoordinates = []

                polygon_coordinates = None

                if('sectionFootprint' in elem['geometry']):
                    polygon_coordinates = elem['geometry']['sectionFootprint'][0] # used for buildings
                else:
                    polygon_coordinates = elem['geometry']['coordinates']

                for i in range(0,int(len(polygon_coordinates)/dimensions)):
                    geometries_coordinates.append(Point(polygon_coordinates[i*dimensions], polygon_coordinates[i*dimensions+1]))
                    ids_coordinates.append(counter_id_coordinates)
                    counter_id_coordinates += 1
                    
                    groupedCoordinates.append((polygon_coordinates[i*dimensions], polygon_coordinates[i*dimensions+1]))

                    if(dimensions == 3 and 'sectionFootprint' not in elem['geometry']): # if it has a 3d representation and it is not a building
                        tridimensional_coordinates.append([polygon_coordinates[i*dimensions], polygon_coordinates[i*dimensions+1], polygon_coordinates[i*dimensions+2]])
                        ids_tridimensional_coordinates.append(counter_id_tridimensional_coordinates)        
                        counter_id_tridimensional_coordinates += 1  

                if('sectionFootprint' in elem['geometry']): # it is a building so a 3d representation must be included (it comes from the coordinates field)
                    for i in range(0,int(len(elem['geometry']['coordinates'])/3)):
                        tridimensional_coordinates.append([elem['geometry']['coordinates'][i*3], elem['geometry']['coordinates'][i*3+1], elem['geometry']['coordinates'][i*3+2]])
                        ids_tridimensional_coordinates.append(counter_id_tridimensional_coordinates)        
                        counter_id_tridimensional_coordinates += 1  

                if(len(groupedCoordinates) >= 3):
                    ids.append(id)
                    geometries.append(Polygon(groupedCoordinates))
        else:
            for i in range(0,int(len(layer_json['coordinates'])/dimensions)):
                
                values_coordinates.append(layer_json['values'][i])
                geometries_coordinates.append(Point(layer_json['coordinates'][i*dimensions], layer_json['coordinates'][i*dimensions+1]))

                if(dimensions == 3):
                    tridimensional_coordinates.append([layer_json['coordinates'][i*dimensions], layer_json['coordinates'][i*dimensions+1], layer_json['coordinates'][i*dimensions+2]])
                    ids_tridimensional_coordinates.append(counter_id_tridimensional_coordinates)        
                    counter_id_tridimensional_coordinates += 1  

        gdf = gpd.GeoDataFrame({'geometry': geometries, 'id': ids}, crs=3395) if not abstract else {}

        gdf_coordinates = gpd.GeoDataFrame({'geometry': geometries_coordinates, 'id': ids_coordinates}, crs=3395) if not abstract else gpd.GeoDataFrame({'geometry': geometries_coordinates, 'value': values_coordinates}, crs=3395)

        df_coordinates3d = None

        if(abstract):
            df_coordinates3d = pd.DataFrame({'geometry': tridimensional_coordinates, 'id': ids_tridimensional_coordinates, 'value': values_coordinates}) if len(tridimensional_coordinates) > 0 and len(ids_tridimensional_coordinates) > 0 else None
        else:
            df_coordinates3d = pd.DataFrame({'geometry': tridimensional_coordinates, 'id': ids_tridimensional_coordinates}) if len(tridimensional_coordinates) > 0 and len(ids_tridimensional_coordinates) > 0 else None

        return {'objects': gdf, 'coordinates': gdf_coordinates, 'coordinates3d': df_coordinates3d}

    def break_into_binary(self, filepath, filename, data, types, dataTypes):

        transformed_data = data

        for index, type in enumerate(types):

            readCoords = 0

            floatList = []

            for i in range(len(transformed_data['data'])):
                geometry = transformed_data['data'][i]['geometry']

                newValue = [readCoords, len(geometry[type])] # where this vector starts and its size

                readCoords += len(geometry[type])

                floatList += geometry[type].copy()

                # geometry[type] = newValue

                transformed_data['data'][i]['geometry'][type] = newValue

            fout = open(os.path.join(filepath,filename+'_'+type+'.data'), 'wb')

            buf = struct.pack(str(len(floatList))+dataTypes[index], *floatList)

            fout.write(buf)
            fout.close()

        json_object = json.dumps(transformed_data)

        with open(os.path.join(filepath,filename+".json"), "w") as outfile:
            outfile.write(json_object)

    def save(self, dir=None, includeGrammar=True):

        if(self.workDir == None and dir == None):
            raise Exception("Directory not specified")

        workDir = None

        if(dir != None):
            # workDir = os.path.join(os.pardir,os.pardir,os.pardir,'public/data',dir)
            workDir = dir
            if(self.workDir == None):
                self.set_work_dir(dir)
        else:
            workDir = self.workDir


        # create dir
        if(os.path.exists(self.workDir) == False):
            os.makedirs(self.workDir)

        grammar_json = {
            "components": [
                {
                    "map": {
                        "camera": self.camera,
                        "knots": [],
                        "interactions": []
                    },
                    "plots": [],
                    "knots": [],
                    "position": {
                        "width": [
                            6,
                            12
                        ],
                        "height": [
                            1,
                            4
                        ]
                    },
                    "widgets": [
                        {
                            "type": "TOGGLE_KNOT"
                        },
                        {
                            "type": "SEARCH"
                        }
                    ],
                    "position": {
                        "width": [
                            6,
                            12
                        ],
                        "height": [
                            1,
                            4
                        ]
                    }
                },
            ],
            "grid": {
                "width": 12,
                "height": 4
            },
            "grammar_position": {
                "width": [
                    1,
                    5
                ],
                "height": [
                    1,
                    4
                ]
            }
        }

        for layer in self.layers['json']:

            grammar_json['components'][0]['knots'].append({"id": "pure"+layer['id'], "integration_scheme": [{"out": {"name": layer['id'], "level": "OBJECTS"}}], "operation": "NONE"})
            grammar_json['components'][0]['map']['knots'].append("pure"+layer['id'])
            grammar_json['components'][0]['map']['interactions'].append("NONE")

            if('data' in layer): # if it is not an abstract layer

                types = []
                dataTypes = []

                if('coordinates' in layer['data'][0]['geometry']):
                    types.append("coordinates")
                    dataTypes.append("d")

                if('normals' in layer['data'][0]['geometry']):
                    types.append("normals")
                    dataTypes.append("f")

                if('indices' in layer['data'][0]['geometry']):
                    types.append("indices")
                    dataTypes.append("I")

                if('ids' in layer['data'][0]['geometry']):
                    types.append("ids")
                    dataTypes.append("I")

                self.break_into_binary(workDir, layer['id'], layer, types, dataTypes)

        if(includeGrammar):
            grammar_json_str = str(json.dumps(grammar_json, indent=4))
            with open(os.path.join(workDir,"grammar.json"), "w", encoding="utf-8") as f:
                f.write(grammar_json_str)

    def view(self):

        website = "http://localhost:5001/"
        # Open url in a new window of the default browser, if possible
        webbrowser.open_new(website)


import requests as internal_requests
from google.oauth2 import id_token
from google.auth.transport import requests

# TODO: pass to env variables
REDIRECT_URI='http://localhost:8080'

class GoogleOAuth:
    def verify_token(self, auth_code):
        token_data = {
            'code': auth_code,
            'client_id': CLIENT_ID,
            'client_secret': CLIENT_SECRET,
            'redirect_uri': REDIRECT_URI,
            'grant_type': 'authorization_code',
        }
        token_url = 'https://oauth2.googleapis.com/token'
        
        token = None
        try:
            token_response = internal_requests.post(token_url, data=token_data)
            token_response_json = token_response.json()
            
            token = token_response_json.get('id_token')
        except Exception as e:
            print(e)
            return None

        print(token)
        try:
            idinfo = id_token.verify_oauth2_token(token, requests.Request(), CLIENT_ID)

            user_data = {
                'email': idinfo.get('email'),
                'name': idinfo.get('name'),
                'picture': idinfo.get('picture'),
                'uid': idinfo.get('sub'),
                'provider': 'google'
            }

            return user_data
        except ValueError as e:
            print(e)
            return None



import os
import pandas as pd
import geopandas as gpd
import numpy as np
import json
import mapbox_earcut as earcut
import struct
from shapely import wkt

from shapely.geometry import Point, Polygon

def break_into_binary(filepath, filename, data, types, dataTypes, type='TRIANGLES_3D_LAYER', renderStyle=['FLAT_COLOR'], styleKey='surface'):

    for index, element in enumerate(types):

        readCoords = 0

        floatList = []

        if('data' in data):
            for i in range(len(data['data'])):
                geometry = data['data'][i]['geometry']

                newValue = [readCoords, len(geometry[element])] # where this vector starts and its size

                readCoords += len(geometry[element])

                floatList += geometry[element].copy()

                geometry[element] = newValue
        else:
            for i in range(len(data)):
                geometry = data[i]['geometry']

                newValue = [readCoords, len(geometry[element])] # where this vector starts and its size

                readCoords += len(geometry[element])

                floatList += geometry[element].copy()

                geometry[element] = newValue

        fout = open(os.path.join(filepath,filename+'_'+element+'.data'), 'wb')

        buf = struct.pack(str(len(floatList))+dataTypes[index], *floatList)

        fout.write(buf)
        fout.close()

        layer = {
            "id": filename,
            "type": type,
            "renderStyle": renderStyle,
            "styleKey": styleKey,
            "data": data
        }

        with open(os.path.join(filepath,filename+".json"), "w") as outfile:
            outfile.write(json.dumps(layer))

'''
    Geometry column must be a string representing a Polygon in the WKT format
'''

def physical_from_csv(filepath, geometry_column='geometry', crs='4326', renderStyle=['FLAT_COLOR'], styleKey='surface'):
    
    df = pd.read_csv(filepath)

    df[geometry_column] = df[geometry_column].apply(wkt.loads)

    gdf = gpd.GeoDataFrame(df, geometry = geometry_column, crs = crs)

    mesh = mesh_from_gdf(gdf)

    directory = os.path.dirname(filepath)
    file_name = os.path.basename(filepath)
    # file name without extension
    file_name_wo_extension = os.path.splitext(file_name)[0]

    types = ["coordinates", "indices"]
    dataTypes = ["d", "I"]
    type = 'TRIANGLES_3D_LAYER'

    if 'indices' not in mesh[0]['geometry'].keys():
        types = [types[0]]
        dataTypes = [dataTypes[0]]
        type = 'POINTS_LAYER'

    break_into_binary(directory, file_name_wo_extension, mesh, types, dataTypes, type, renderStyle, styleKey)

def physical_from_geojson(filepath, bbox = None, renderStyle=['FLAT_COLOR'], styleKey='surface'):

    gdf = gpd.read_file(filepath)

    if(bbox != None):
        gdf = gdf.cx[bbox[0]:bbox[2], bbox[1]:bbox[3]]

    mesh = mesh_from_gdf(gdf)

    directory = os.path.dirname(filepath)
    file_name = os.path.basename(filepath)
    # file name without extension
    file_name_wo_extension = os.path.splitext(file_name)[0]

    break_into_binary(directory, file_name_wo_extension, mesh, ["coordinates", "indices"], ["d", "I"], 'TRIANGLES_3D_LAYER', renderStyle, styleKey)

'''
    Geometry has to be Polygon or Multipolygon
'''
def mesh_from_gdf(gdf):

    gdf_transformed = gdf.to_crs(3395)

    mesh = []

    for geometry in gdf_transformed.geometry:
        if geometry.geom_type == 'Point':
            nodes_3d = [geometry.x, geometry.y, 0]
            mesh.append({'geometry': {'coordinates': [round(item,4) for item in nodes_3d]}})

        else:
            sub_geometries = []
            if geometry.geom_type == 'MultiPolygon':
                sub_geometries = list(geometry)
            elif geometry.geom_type == 'Polygon':
                sub_geometries = [geometry]

            for element in sub_geometries:

                x, y = element.exterior.coords.xy

                nodes = list(zip(x,y))
                rings = [len(nodes)]

                indices = earcut.triangulate_float64(nodes, rings)

                nodes = np.array(nodes)

                nodes = nodes.flatten().tolist()
                indices = indices.tolist()

                nodes_3d = []

                for i in range(int(len(nodes)/2)):
                    nodes_3d.append(nodes[i*2])
                    nodes_3d.append(nodes[i*2+1])
                    nodes_3d.append(0)

                mesh.append({'geometry': {'coordinates': [round(item,4) for item in nodes_3d], 'indices': indices}})

    return mesh

'''
    Generate mesh json file based on shapefile
'''
def physical_from_shapefile(filepath, layerName, bpoly=None, isBbox = False, renderStyle=['FLAT_COLOR'], styleKey='surface'):
    '''
        In the same folder as the .shp file there must be a .prj and .shx files   

        The bounding box must be in the 4326 projection

        Only works for 2D geometries

        Returns gdf in 3395
    '''

    loaded_shp = []

    if(isBbox):
        bbox_series_4326 = gpd.GeoSeries([Point(bpoly[1], bpoly[0]), Point(bpoly[3], bpoly[2])], crs=4326)
        
        loaded_shp = gpd.read_file(filepath, bbox=bbox_series_4326)

        bbox_series_4326 = bbox_series_4326.to_crs(3395)

        loaded_shp = loaded_shp.to_crs(3395)
        loaded_shp = loaded_shp.clip([bbox_series_4326[0].x, bbox_series_4326[0].y, bbox_series_4326[1].x, bbox_series_4326[1].y])
    else:

        loaded_shp = gpd.read_file(filepath)
        loaded_shp = loaded_shp.to_crs(3395)

        if(bpoly != None):

            bpoly_series_4326 = gpd.GeoSeries([Polygon(bpoly)], crs=4326)
            bpoly_series_4326 = bpoly_series_4326.to_crs(3395)

            loaded_shp = loaded_shp.clip(bpoly_series_4326)

    zip_code_coordinates = []

    data = []
    objectId = []
    coordinates_geometries = []
    coordinates_ids = []
    coord_id_counter = 0

    for id, row in enumerate(loaded_shp.iloc):

        objectId.append(id)

        geometries = []
        if row['geometry'].geom_type == 'MultiPolygon':
            geometries = list(row['geometry'])
        elif row['geometry'].geom_type == 'Polygon':
            geometries = [row['geometry']]

        coordinates = []
        indices = []
        count = 0

        for geometry in geometries:
            points = np.array(geometry.exterior.coords[0:-1]) # remove last one (repeated)
            rings = np.array([len(points)])

            ind = earcut.triangulate_float64(points, rings)
            ind = (ind+count).tolist()
            indices += ind

            points = points.flatten().tolist()

            for i in range(0, len(points), 2):
                coordinates.append(points[i])
                coordinates.append(points[i+1])
                coordinates_geometries.append(Point(points[i], points[i+1]))
                coordinates_ids.append(coord_id_counter)
                coord_id_counter += 1
                coordinates.append(0)

            count = int(len(coordinates)/3)
        
        zip_code_coordinates += coordinates

        data.append({
            "geometry": {
                "coordinates": [round(item,4) for item in coordinates.copy()],
                "indices": indices.copy()
            }
        })

    outputfile = os.path.join(os.path.dirname(filepath), layerName+'.json') 

    with open(outputfile, "w", encoding="utf-8") as f:
        
        # result = {
        #     "id": layerName,
        #     "type": type,
        #     "renderStyle": renderStyle,
        #     "styleKey": styleKey,
        #     "data": data
        # }

        types = []
        dataTypes = []

        if('coordinates' in data[0]['geometry']):
            types.append("coordinates")
            dataTypes.append("d")

        if('normals' in data[0]['geometry']):
            types.append("normals")
            dataTypes.append("f")

        if('indices' in data[0]['geometry']):
            types.append("indices")
            dataTypes.append("I")

        if('ids' in data[0]['geometry']):
            types.append("ids")
            dataTypes.append("I")

        break_into_binary(os.path.dirname(filepath), layerName, data, types, dataTypes, 'TRIANGLES_3D_LAYER', renderStyle, styleKey)

        # layer_json_str = str(json.dumps(result))
        # f.write(layer_json_str)

    loaded_shp['id'] = objectId

    coordinates_gdf = gpd.GeoDataFrame({'geometry': coordinates_geometries, "id": coordinates_ids}, crs=3395)

    return {'objects': loaded_shp, 'coordinates': coordinates_gdf, 'coordinates3d': None}


import sys
import numpy as np
import xarray as xr

from netCDF4 import Dataset
from termcolor import colored
# from wrf import getvar, interplevel, to_np, ALL_TIMES
from pyproj import Transformer

# class WRFOutputReader(object):
#     def __init__(self) -> None:
#         self.__className = "WRFOutputReader"
#         self.__variablesList = ('PM10', 'so2', 'co', 'o3', 'PM2_5_DRY', 'no', 'no2', 'no3', 'T2', 'rh', 'wind10m', 'accRain', '3hRain', '24hRain')
#         self.resetClass()
        
#     def __processRain(self, var_key, time_steps):
#         acc = self.__nTimes - 1
    
#         if var_key == '24hRain':
#             acc = 24

#         elif var_key == '3hRain':
#             acc = 3
        
#         data = []

#         rainc = self.__ncDs.variables['RAINC'][:]
#         rainnc = self.__ncDs.variables['RAINNC'][:]

#         rainAc = []

#         _t0 = 0 if self.__startTstepIdx == 0 else self.__startTstepIdx - acc

#         for t in range(_t0, self.__endTstepIdx + 1):
#             tRainc = rainc[t]
#             tRainnc = rainnc[t]

#             rainAc.append(np.flipud(tRainc + tRainnc))

#         self.__accPrec = rainAc
#         return data
    
#     def __process4DVariable(self, var_key, level, time_steps):
#         data = []

#         for t in time_steps:
#             var_time_matrix = getvar(self.__ncData, var_key, timeidx=t,
#                         squeeze=False, meta=False)
#             pressure_time_matrix = getvar(self.__ncData, 'pressure', timeidx=t,
#                         squeeze=False, meta=False)

#             var_time_matrix = to_np(interplevel(var_time_matrix, pressure_time_matrix, level))

#             if np.ma.isMaskedArray(var_time_matrix):
#                 var_time_matrix = list(var_time_matrix)
#                 var_time_matrix.reverse()
#                 data.append(var_time_matrix)

#         data = np.array(data)
#         return data
    
#     def getNTimes(self):
#         return self.__nTimes
    
#     def getVariableData(self, var_key, var_level=None, time_steps=[]):
#         active_var = None
#         data = None
        

#         if len(time_steps) == 0: time_steps = range(self.__nTimes)

#         # Set active_var
#         for i in range(len(self.__variablesList)):
#             if var_key.lower() == self.__variablesList[i].lower(): 
#                 active_var = self.__variablesList[i]
#                 break
        
#         # Set data according to the variable
#         if active_var:
#             if 'Rain' in active_var:
#                 self.__processRain(active_var, time_steps)

#             elif active_var == 'T2':
#                 data = self.__ncData.variables[active_var][:]
#                 data -= 273 # Kelvin to Celsius
            
#             elif active_var == 'wind10m':
#                 # to do
#                 data = []

#             elif self.__ncData.variables[active_var].shape == 4:
#                 self.__process4DVariable(active_var, var_level, time_steps)
                    
#         else:
#             sys.exit()
              
#         return data
       
#     def getLatLon(self):
#         # return self.__latMatrix, self.__lonMatrix
#         return self.__latList, self.__lonList
    
#     def setNcData(self, ncFilePath):
#         try:
#             self.__ncData = Dataset(ncFilePath)

#         except Exception as error:
#             msg = f"[{self.__className} - setNcData]\n{error}"
#             print(colored(f"{msg}", 'red'))
#             sys.exit()

#     def resetClass(self):
#         self.__ncData = None
#         self.__gridId = None
#         self.__latList = None
#         self.__lonList = None
#         self.__nTimes = None
    
#     def setAttributes(self):
#         self.__latMatrix = sorted(self.__ncData.variables['XLAT'][:][0], key=lambda x: x[0], reverse=True)
#         self.__lonMatrix = self.__ncData.variables['XLONG'][:][0]
        
#         self.__latList = np.array([row[0] for row in self.__latMatrix] )
#         self.__lonList = np.array(self.__lonMatrix[0])

#         self.__nRows = self.__ncData.dimensions['south_north'].size
#         self.__nCols = self.__ncData.dimensions['west_east'].size

#         self.__nTimes = self.__ncData.dimensions['Time'].size

# def my_function(model, file_path, target_variables, time_steps=[], bbox=[]):

#     ncReader = None

#     if model.lower() == 'wrf':
#         ncReader = WRFOutputReader()
    
#     if ncReader:
#         ncReader.setNcData(file_path)
#         ncReader.setAttributes()

#         for var_obj in target_variables:
#             var_key   = var_obj.variable
#             var_level = var_obj.level

#             data = ncReader.getVariableData(var_key, var_level, time_steps)


    
# if __name__ == '__main__':
#     # path = f'../../examples/wrf_chicago'
#     # startDate = "2016-07-01"
#     # gId = "2"

#     # filepath = f"{path}/wrfout_d0{gId}_{startDate}.nc"

#     # thematic_from_wrf(filepath, ['T2'], 4326, range(1, 5), [33, -94, 42, -81])

#     path = f'../../examples/wrf_chicago'
#     startDate = "2023-06-09_20_00_00"
#     gId = "3"

#     filepath = f"{path}/wrfout_d0{gId}_{startDate}"

#     # thematic_from_wrf(filepath, ['T2'], 4326, range(1, 5), [33, -94, 42, -81])

#     obj1 = {'variable': 'T2', 'level': None}
#     obj2 = {'variable': 'rh', 'level': 850}
#     varList = [obj1, obj2]

#     my_function('wrf', filepath, varList)

#     # for v in ('PM10', 'so2', 'co', 'o3', 'PM2_5_DRY', 'no', 'no2', 'no3', 'T2'):
#     #     myvar = wrfout.getVariableData(v)
#     # print(myvar)




from flask import request, abort, jsonify, g
import requests
import json
import sqlite3
from utk_curio.backend.extensions import db
from utk_curio.backend.app.users.models import User, UserSession
from utk_curio.backend.app.services.google_oauth import GoogleOAuth
from utk_curio.backend.app.middlewares import require_auth
import uuid
import os
import zlib
import time
import hashlib
import mmap
import ast

# The Flask app
from utk_curio.backend.app.api import bp


# Sandbox address
api_address='http://'+os.getenv('FLASK_SANDBOX_HOST', 'localhost')
api_port=int(os.getenv('FLASK_SANDBOX_PORT', 2000))


inputTypesSupported = {
    "DATA_LOADING": [],
    "DATA_EXPORT": ["DATAFRAME", "GEODATAFRAME"],
    "DATA_CLEANING": ["DATAFRAME", "GEODATAFRAME"],
    "DATA_TRANSFORMATION": ["DATAFRAME", "GEODATAFRAME"],
    "COMPUTATION_ANALYSIS": ["DATAFRAME", "GEODATAFRAME", "VALUE", "LIST", "JSON"],
    "FLOW_SWITCH": ["DATAFRAME", "GEODATAFRAME", "VALUE", "LIST", "JSON"],
    "VIS_UTK": ["GEODATAFRAME"],
    "VIS_VEGA": ["DATAFRAME"],
    "VIS_TABLE": ["DATAFRAME", "GEODATAFRAME"],
    "VIS_TEXT": ["VALUE"],
    "VIS_IMAGE": ["LIST"],
    "CONSTANTS": [],
    "DATA_POOL": ["DATAFRAME", "GEODATAFRAME"],
    "MERGE_FLOW": ["DATAFRAME", "GEODATAFRAME", "VALUE", "LIST", "JSON"],
}

outputTypesSupported = {
    "DATA_LOADING": ["DATAFRAME", "GEODATAFRAME"],
    "DATA_EXPORT": [],
    "DATA_CLEANING": ["DATAFRAME", "GEODATAFRAME"],
    "DATA_TRANSFORMATION": ["DATAFRAME", "GEODATAFRAME"],
    "COMPUTATION_ANALYSIS": ["DATAFRAME", "GEODATAFRAME", "VALUE", "LIST", "JSON"],
    "FLOW_SWITCH": ["DATAFRAME", "GEODATAFRAME", "VALUE", "LIST", "JSON"],
    "VIS_UTK": ["GEODATAFRAME"],
    "VIS_VEGA": ["DATAFRAME"],
    "VIS_TABLE": ["DATAFRAME", "GEODATAFRAME"],
    "VIS_TEXT": ["VALUE"],
    "VIS_IMAGE": ["LIST"],
    "CONSTANTS": ["VALUE"],
    "DATA_POOL": ["DATAFRAME", "GEODATAFRAME"],
    "MERGE_FLOW": ["DATAFRAME", "GEODATAFRAME", "VALUE", "LIST", "JSON"],
}

attributeIds = {
    "DATAFRAME": "1",
    "GEODATAFRAME": "2",
    "VALUE": "3",
    "LIST": "4",
    "JSON": "5"
}

TYPE_MAP = {
    "computation_analysis": "COMPUTATION_ANALYSIS",
    "data_cleaning": "DATA_CLEANING",
    "data_export": "DATA_EXPORT",
    "data_loading": "DATA_LOADING",
    "data_transformation": "DATA_TRANSFORMATION",
    "utk": "VIS_UTK",
    "vega_lite": "VIS_VEGA"
}

FOLDER_MAP = {
    "COMPUTATION_ANALYSIS": "computation_analysis",
    "DATA_CLEANING": "data_cleaning",
    "DATA_EXPORT": "data_export",
    "DATA_LOADING": "data_loading",
    "DATA_TRANSFORMATION": "data_transformation",
    "VIS_UTK": "utk",
    "VIS_VEGA": "vega_lite"
}

TEMPLATE_DIR = "./templates"

def get_db_path():
    launch_dir = os.environ.get("CURIO_LAUNCH_CWD", os.getcwd())
    db_path = os.path.join(launch_dir, ".curio", "provenance.db")
    return db_path

@bp.after_request
def add_cors_headers(response):
    response.headers.add('Access-Control-Allow-Origin', '*')
    response.headers.add('Access-Control-Allow-Headers', 'Content-Type,Authorization')
    response.headers.add('Access-Control-Allow-Methods', 'GET,PUT,POST,DELETE')
    return response

@bp.route('/')
def root():
    abort(403)

@bp.route('/live')
def live():
    return 'Backend is live.'

@bp.route('/upload', methods=['POST'])
def upload_file():

    if 'file' not in request.files:
        return 'No file part'

    file = request.files['file']

    if file.filename == '':
        return 'No selected file'
    
    response = requests.post(api_address+":"+str(api_port)+"/upload", files={'file': file}, data={'fileName': file.filename})

    if response.status_code == 200:
        return 'File uploaded successfully'
    else:
        return 'Error uploading file'


def transform_to_vega(data):
    """
    Transforms a pandas-style JSON (column-based) to Vega-Lite-ready JSON (row-based).
    
    Args:
        data (dict): The original pandas-style JSON data.

    Returns:
        dict: The transformed Vega-Lite-ready JSON data.
    """
    if "data" in data and isinstance(data["data"], dict):
        columns = list(data["data"].keys())
        values = []

        # Assuming all columns have the same number of rows
        num_rows = len(data["data"][columns[0]])

        for i in range(num_rows):
            row = {col: data["data"][col][i] for col in columns}
            values.append(row)

        return values

    return data

@bp.route('/get', methods=['GET'])
def get_file():
    file_path = request.args.get('fileName')
    vega = request.args.get('vega', 'false').lower() == 'true'

    if not file_path:
        return 'No file name specified', 400

    if not os.path.exists(file_path):
        return 'File does not exist', 404

    try:
        # Using mmap for efficient memory-mapped loading
        with open(file_path, "rb") as file:
            with mmap.mmap(file.fileno(), 0, access=mmap.ACCESS_READ) as mmapped_file:
                # Decompress and decode directly from the memory-mapped file
                decompressed_data = zlib.decompress(mmapped_file[:])
                data = json.loads(decompressed_data.decode('utf-8'))

                if isinstance(data, str):
                    data = json.loads(data)
                
                if vega:
                    data = transform_to_vega(data)

        return jsonify(data)

    except Exception as e:
        return f'Error loading file: {str(e)}', 500

    return jsonify(data), 200

# def save_memory_mapped_file(data, shared_disk_path='./data/'):
#     """
#     Saves the input data as a memory-mapped JSON file with a unique name.

#     Args:
#         input_data (dict): The data to be saved.
#         shared_disk_path (str): Path to the directory for saving the file.

#     Returns:
#         str: The path of the saved memory-mapped file.
#     """
#     # Ensure the shared directory exists
#     os.makedirs(shared_disk_path, exist_ok=True)
    
#     # Create a unique filename using hash of the input and current time
#     input_hash = hashlib.sha256(json.dumps(data).encode('utf-8')).hexdigest()
#     timestamp = str(int(time.time()))
#     unique_filename = f"{timestamp}_{input_hash[:10]}.json"  # Using .json for clarity
#     file_path = os.path.join(shared_disk_path, unique_filename)

#     # Save the input data directly without compression as a memory-mapped file
#     compressed_data = zlib.compress(json.dumps(data).encode('utf-8'))
#     with open(file_path, "wb") as file:
#         file.write(compressed_data)

#     return file_path

@bp.route('/processPythonCode', methods=['POST'])
def process_python_code():

    # response = requests.post(api_address+":"+str(api_port)+"/api/v2/execute",
    #     data=json.dumps({
    #         "language": "python",
    #         "version": "3.12.0",
    #         "files": [
    #             {
    #                 "name": "test.py",
    #                 "content": request.json['code']
    #             }
    #         ]
    #     }),
    #     headers={"Content-Type": "application/json"},
    # )
    # print(api_address, str(api_port))
    # print(api_address+":"+str(api_port))
    # print(request.json, flush=True)

    code = request.json['code']
    boxType = request.json['boxType']
    input = {'path': "", 'dataType': ""}
    if(request.json['input']):

        if(request.json['input']['dataType'] == 'outputs'):
            input['path'] = request.json['input']['data']
            input['dataType'] = 'outputs'
        else:
            input['path'] = request.json['input']['path']
            input['dataType'] = request.json['input']['dataType']

    # print("-------------- input_data", input_data, flush=True)

    # file_path = input_data # save_memory_mapped_file(input_data)
    # if(file_path != ""):
    #     with open(file_path, "rb") as file:
    #         with mmap.mmap(file.fileno(), 0, access=mmap.ACCESS_READ) as mmapped_file:
    #             # Decompress and load the data directly from memory
    #             input = json.loads(zlib.decompress(mmapped_file).decode('utf-8'))
    #             print("-------------- input", input, flush=True)
    try:
        # print('-----------', input['path'], boxType, flush=True)
        response = requests.post(api_address+":"+str(api_port)+"/exec",
                                data=json.dumps({
                                    "code": code,
                                    "file_path": input['path'],
                                    "boxType": boxType,
                                    "dataType": input['dataType']
                                }),
                                headers={"Content-Type": "application/json"},
                                )
        
        try:
            # print("-------------- response", response.json(), flush=True)
            response = response.json()
            stdout = response['stdout']
            stderr = response['stderr']
            output = response['output'] # contains path and dataType
            print(output, flush=True)
            # print("-------------- response", {'stdout': stdout, 'stderr': stderr, 'inputTypes': input_types, 'output': output_path, 'outputType': output_type}, flush=True)
            
            return {'stdout': stdout, 'stderr': stderr, 'input': input, 'output': output}
        finally:
            pass
    #         if os.path.exists(output_path):
    #             os.remove(output_path)
    finally:
        pass
    #     if os.path.exists(file_path):
    #         os.remove(file_path)

@bp.route('/toLayers', methods=['POST'])
def toLayers():

    if(request.json['geojsons'] == None):
        abort(400, "geojsons were not included in the post request")

    response = requests.post(api_address+":"+str(api_port)+"/toLayers",
                             data=json.dumps({
                                 "geojsons": request.json['geojsons']
                             }),
                             headers={"Content-Type": "application/json"},
                             )

    return response.json()

@bp.route('/signin', methods=['POST'])
def signin():
    # google_oauth = GoogleOAuth()
    # user_data = google_oauth.verify_token(request.json.get('token'))
    # if not user_data:
    #     return jsonify({'error': 'Invalid token'}), 400

    # create new session token
    # new_session = UserSession(user_id=user.id)
    user_data = {
        'id': 1,
        'name': 'Test',
        'email': 'Test@mail.com',
        'provider': "",
        'uid': "",
        'picture': "",
        'type': 'programmer'
    }
    new_session = UserSession(user_id=user_data.get('id'))
    db.session.add(new_session)
    db.session.commit()


    # get user from database

    user = User.query.filter_by(
        id=user_data.get('id'),
        # provider=user_data.get('provider'),
        # provider_uid= user_data.get('uid')
    ).first()

    if user:
        user.name = user_data.get('name')
        user.profile_image = user_data.get('picture')
    else:
        user = User(
            email=user_data.get('email'),
            name=user_data.get('name'),
            profile_image=user_data.get('picture'),
            provider=user_data.get('provider'),
            provider_uid=user_data.get('uid'))
        db.session.add(user)
    db.session.commit()


    return jsonify({
        'user': {
            'name': user.name,
            'profile_image': user.profile_image,
            'type': user.type
        },
        'token': new_session.token
    }), 200

@bp.route('/getUser', methods=['GET'])
@require_auth
def get_user():
    user = g.user
    return jsonify({
        'user': {
            'name': user.name,
            'profile_image': user.profile_image,
            'type': user.type
        }
    }), 200

@bp.route('/saveUserType', methods=['POST'])
@require_auth
def save_user_type():
    new_type = request.json.get('type')
    user = g.user
    user.type = new_type
    db.session.commit()

    return jsonify({
        'user': {
            'name': user.name,
            'profile_image': user.profile_image,
            'type': user.type
        }
    }), 200

@bp.route('/saveUserProv', methods=['POST'])
def save_user_prov(): # only save if user with that name does not exist on the database

    # conn = sqlite3.connect('backend/provenance.db')
    # db_path = os.path.join(os.getcwd(), ".curio", "provenance.db")
    db_path = get_db_path()
    conn = sqlite3.connect(db_path)
    cursor = conn.cursor()

    user = request.json.get('user')

    # Check if the user_name already exists in the table
    cursor.execute('''SELECT 1 FROM user WHERE user_name = ?''', (user['user_name'],))
    existing_record = cursor.fetchone()

    if not existing_record:
        data_to_insert = (user['user_name'], user['user_type'], user['user_IP'])
        cursor.execute('''INSERT INTO user (user_name, user_type, user_IP)
                        VALUES (?, ?, ?)''', data_to_insert)

    conn.commit()
    conn.close()

    return "",200

@bp.route('/saveWorkflowProv', methods=['POST'])
def save_workflow_prov():

    # db_path = os.path.join(os.getcwd(), ".curio", "provenance.db")
    db_path = get_db_path()
    conn = sqlite3.connect(db_path)
    cursor = conn.cursor()

    workflow_name = request.json.get('workflow')

    # starting a new version counter for the workflow
    cursor.execute('''INSERT INTO version (version_number)
                    VALUES (?)''', ('1.0',))

    conn.commit()

    # id of the new just added version element
    version_id = cursor.lastrowid

    # creating new versioned element for the new workflow
    cursor.execute('''INSERT INTO versionedElement (version_id)
                    VALUES (?)''', (version_id,))

    conn.commit()

    # id of the new just added versioned element
    ve_id = cursor.lastrowid

    # creating new workflow
    cursor.execute('''INSERT INTO workflow (workflow_name, ve_id)
                    VALUES (?, ?)''', (workflow_name, ve_id,))

    conn.commit()
    conn.close()

    return "",200

@bp.route('/newBoxProv', methods=['POST'])
def new_box_prov():

    # db_path = os.path.join(os.getcwd(), ".curio", "provenance.db")
    db_path = get_db_path()
    conn = sqlite3.connect(db_path)
    cursor = conn.cursor()

    # // new version (increment version number based on previous old workflow that points to a ve that points to the version)
    # // new versioned element (pointing to the versioned element of the old workflow and pointing to the new version)
    # // new workflow
    # // point new workflow to the new versioned element
    # // create relation for new activity (output relation)
    # // new activity (pointing to the new workflow and to the new relation as output)
    # // duplicate all activities that point to the old workflow and point to the new one (duplicate relations tied to activities)

    data = request.json.get('data')

    # last workflow created with this name
    cursor.execute("SELECT * FROM workflow WHERE workflow_id = (SELECT MAX(workflow_id) FROM workflow WHERE workflow_name = ?)", (data['workflow_name'],))
    old_workflow = cursor.fetchone()

    # getting versionedElement attached to the old workflow
    cursor.execute("SELECT * FROM versionedElement WHERE ve_id = ?", (old_workflow[2],))
    old_workflow_ve = cursor.fetchone()

    # getting the version atteched to the old workflow
    cursor.execute("SELECT * FROM version WHERE version_id = ?", (old_workflow_ve[1],))
    version = cursor.fetchone()

    new_version_number = str(float(version[1])+1.0)

    # creating new version
    cursor.execute('''INSERT INTO version (version_number)
                    VALUES (?)''', (new_version_number,))

    conn.commit()

    # id of the new just added version
    version_id = cursor.lastrowid

    # creating new versioned element for the new workflow
    cursor.execute('''INSERT INTO versionedElement (previous_ve_id, version_id)
                    VALUES (?, ?)''', (old_workflow_ve[0], version_id,))

    conn.commit()

    # id of the new just added versioned element
    ve_id = cursor.lastrowid

    # creating new workflow
    cursor.execute('''INSERT INTO workflow (workflow_name, ve_id)
                    VALUES (?, ?)''', (data['workflow_name'], ve_id,))

    conn.commit()

    # id of the new just added workflow
    workflow_id = cursor.lastrowid

    # creating relation for new activity
    cursor.execute('''INSERT INTO relation (relation_name)
                VALUES (?)''', (data['activity_name']+"_"+"out",))

    conn.commit()

    # id of the new just added relation
    output_relation_id = cursor.lastrowid

    boxType = data['activity_name'].split("-")[0]

    # adding a attributeRelation to each output type that this activity supports
    for outputType in outputTypesSupported[boxType]:
        cursor.execute('''INSERT INTO attributeRelation (attribute_id, relation_id)
            VALUES (?, ?)''', (attributeIds[outputType], output_relation_id,))

    # creating new activity
    cursor.execute('''INSERT INTO activity (workflow_id, activity_name, output_relation_id)
                    VALUES (?, ?, ?)''', (workflow_id, data['activity_name'], output_relation_id,))

    conn.commit()

    # getting all activities that point to the old workflow
    cursor.execute("SELECT activity_name, input_relation_id, output_relation_id, ve_id FROM activity WHERE workflow_id = ?", (old_workflow[0],))

    activities = cursor.fetchall()

    duplicated_output_relations = {} # dict of old to new ids.

    duplicated_activities_ids = []

    # duplicating all activities and making them point to the new workflow
    for activity in activities:

        # getting the old output relation of duplicated activity
        cursor.execute("SELECT relation_id, relation_name FROM relation WHERE relation_id = ?", (activity[2],))
        old_output_relation = cursor.fetchone()

        # # duplicate the old output relation
        # cursor.execute('''INSERT INTO relation (relation_name)
        #             VALUES (?)''', (old_output_relation[1],))

        # conn.commit()

        # # id of the new just added relation
        # output_relation_id = cursor.lastrowid

        boxType = activity[0].split("-")[0]

        # # adding a attributeRelation to each output type that this activity supports
        # for outputType in outputTypesSupported[boxType]:
        #     cursor.execute('''INSERT INTO attributeRelation (attribute_id, relation_id)
        #         VALUES (?, ?)''', (attributeIds[outputType], output_relation_id,))

        # duplicated_output_relations[old_output_relation[0]] = output_relation_id # mapping old to new ids

        # cursor.execute('''INSERT INTO activity (workflow_id, activity_name, input_relation_id, output_relation_id, ve_id)
        #         VALUES (?, ?, ?, ?, ?)''', (workflow_id, activity[0], activity[1], output_relation_id, activity[3],))

        cursor.execute('''INSERT INTO activity (workflow_id, activity_name, input_relation_id, output_relation_id, ve_id)
                VALUES (?, ?, ?, ?, ?)''', (workflow_id, activity[0], activity[1], old_output_relation[0], activity[3],))

        conn.commit()

        # id of the new just added activity
        activity_id = cursor.lastrowid

        duplicated_activities_ids.append(activity_id)

    # # updating duplicated activities to point to the duplicated relations
    # for old_output_id in duplicated_output_relations:
    #     cursor.execute("SELECT activity_id FROM activity WHERE input_relation_id = ?", (old_output_id,))
    #     activities = cursor.fetchall()

    #     for activity in activities:
    #         if activity[0] in duplicated_activities_ids: # this is a duplicated activity that needs to have input field updated to point to new duplicated relation
    #             cursor.execute("UPDATE activity SET input_relation_id = ? WHERE activity_id = ?", (duplicated_output_relations[old_output_id], activity[0],))
    #             conn.commit()

    conn.commit()
    conn.close()

    # // TODO: new and duplicated activities can also be versioned by creating a new versioned element
    return "",200

@bp.route('/deleteBoxProv', methods=['POST'])
def delete_box_prov():

    # // new version (increment version number based on previous old workflow that points to a ve that points to the version)
    # // new versioned element (pointing to the versioned element of the old workflow and pointing to the new version)
    # // new workflow
    # // point new workflow to the new versioned element
    # // duplicate all activities (but the excluded box) that point to the old workflow and point to the new one

    # db_path = os.path.join(os.getcwd(), ".curio", "provenance.db")
    db_path = get_db_path()
    conn = sqlite3.connect(db_path)
    cursor = conn.cursor()

    data = request.json.get('data')

    # last workflow created with this name
    cursor.execute("SELECT * FROM workflow WHERE workflow_id = (SELECT MAX(workflow_id) FROM workflow WHERE workflow_name = ?)", (data['workflow_name'],))
    old_workflow = cursor.fetchone()

    # getting versionedElement attached to the old workflow
    cursor.execute("SELECT * FROM versionedElement WHERE ve_id = ?", (old_workflow[2],))
    old_workflow_ve = cursor.fetchone()

    # getting the version atteched to the old workflow
    cursor.execute("SELECT * FROM version WHERE version_id = ?", (old_workflow_ve[1],))
    version = cursor.fetchone()

    new_version_number = str(float(version[1])+1.0)

    # creating new version
    cursor.execute('''INSERT INTO version (version_number)
                    VALUES (?)''', (new_version_number,))

    conn.commit()

    # id of the new just added version
    version_id = cursor.lastrowid

    # creating new versioned element for the new workflow
    cursor.execute('''INSERT INTO versionedElement (previous_ve_id, version_id)
                    VALUES (?, ?)''', (old_workflow_ve[0], version_id,))

    conn.commit()

    # id of the new just added versioned element
    ve_id = cursor.lastrowid

    # creating new workflow
    cursor.execute('''INSERT INTO workflow (workflow_name, ve_id)
                    VALUES (?, ?)''', (data['workflow_name'], ve_id,))

    conn.commit()

    # id of the new just added workflow
    workflow_id = cursor.lastrowid

    # getting all activities that point to the old workflow
    cursor.execute("SELECT activity_name, input_relation_id, output_relation_id, ve_id FROM activity WHERE workflow_id = ?", (old_workflow[0],))

    activities = cursor.fetchall()

    duplicated_output_relations = {} # dict of old to new ids.

    duplicated_activities_ids = []

    # duplicating all activities (except deleted one) and making them point to the new workflow
    for activity in activities:
        if(activity[0] != data['activity_name']):

            # getting the old output relation of duplicated activity
            cursor.execute("SELECT relation_id, relation_name FROM relation WHERE relation_id = ?", (activity[2],))
            old_output_relation = cursor.fetchone()

            # # duplicate the old output relation
            # cursor.execute('''INSERT INTO relation (relation_name)
            #             VALUES (?)''', (old_output_relation[1],))

            # conn.commit()

            # # id of the new just added relation
            # output_relation_id = cursor.lastrowid

            boxType = activity[0].split("-")[0]

            # # adding a attributeRelation to each output type that this activity supports
            # for outputType in outputTypesSupported[boxType]:
            #     cursor.execute('''INSERT INTO attributeRelation (attribute_id, relation_id)
            #         VALUES (?, ?)''', (attributeIds[outputType], output_relation_id,))

            # duplicated_output_relations[old_output_relation[0]] = output_relation_id # mapping old to new ids

            # cursor.execute('''INSERT INTO activity (workflow_id, activity_name, input_relation_id, output_relation_id, ve_id)
            #         VALUES (?, ?, ?, ?, ?)''', (workflow_id, activity[0], activity[1], output_relation_id, activity[3],))

            cursor.execute('''INSERT INTO activity (workflow_id, activity_name, input_relation_id, output_relation_id, ve_id)
                    VALUES (?, ?, ?, ?, ?)''', (workflow_id, activity[0], activity[1], old_output_relation[0], activity[3],))

            conn.commit()

            # id of the new just added activity
            activity_id = cursor.lastrowid

            duplicated_activities_ids.append(activity_id)

    # # updating duplicated activities to point to the duplicated relations
    # for old_output_id in duplicated_output_relations:
    #     cursor.execute("SELECT activity_id FROM activity WHERE input_relation_id = ?", (old_output_id,))
    #     activities = cursor.fetchall()

    #     for activity in activities:
    #         if activity[0] in duplicated_activities_ids: # this is a duplicated activity that needs to have input field updated to point to new duplicated relation
    #             cursor.execute("UPDATE activity SET input_relation_id = ? WHERE activity_id = ?", (duplicated_output_relations[old_output_id], activity[0],))
    #             conn.commit()

    conn.commit()
    conn.close()

    # // TODO: new and duplicated activities can also be versioned by creating a new versioned element

    return "",200

@bp.route('/newConnectionProv', methods=['POST'])
def new_connection_prov():

    # db_path = os.path.join(os.getcwd(), ".curio", "provenance.db")
    db_path = get_db_path()
    conn = sqlite3.connect(db_path)
    cursor = conn.cursor()

    data = request.json.get('data')

    # // new version (increment version number based on previous old workflow that points to a ve that points to the version)
    # // new versioned element (pointing to the versioned element of the old workflow and pointing to the new version)
    # // new workflow
    # // point new workflow to the new versioned element
    # // duplicate all activities that point to the old workflow and point to the new one (duplicate relations tied to activities)
    # // update input relation of the activity

    # last workflow created with this name
    cursor.execute("SELECT * FROM workflow WHERE workflow_id = (SELECT MAX(workflow_id) FROM workflow WHERE workflow_name = ?)", (data['workflow_name'],))
    old_workflow = cursor.fetchone()

    # getting versionedElement attached to the old workflow
    cursor.execute("SELECT * FROM versionedElement WHERE ve_id = ?", (old_workflow[2],))
    old_workflow_ve = cursor.fetchone()

    # getting the version atteched to the old workflow
    cursor.execute("SELECT * FROM version WHERE version_id = ?", (old_workflow_ve[1],))
    version = cursor.fetchone()

    new_version_number = str(float(version[1])+1.0)

    # creating new version
    cursor.execute('''INSERT INTO version (version_number)
                    VALUES (?)''', (new_version_number,))

    conn.commit()

    # id of the new just added version
    version_id = cursor.lastrowid

    # creating new versioned element for the new workflow
    cursor.execute('''INSERT INTO versionedElement (previous_ve_id, version_id)
                    VALUES (?, ?)''', (old_workflow_ve[0], version_id,))

    conn.commit()

    # id of the new just added versioned element
    ve_id = cursor.lastrowid

    # creating new workflow
    cursor.execute('''INSERT INTO workflow (workflow_name, ve_id)
                    VALUES (?, ?)''', (data['workflow_name'], ve_id,))

    conn.commit()

    # id of the new just added workflow
    workflow_id = cursor.lastrowid

    # getting all activities that point to the old workflow
    cursor.execute("SELECT activity_name, input_relation_id, output_relation_id, ve_id FROM activity WHERE workflow_id = ?", (old_workflow[0],))

    activities = cursor.fetchall()

    duplicated_output_relations = {} # dict of old to new ids.

    duplicated_activities_ids = []

    # duplicating all activities and making them point to the new workflow
    for activity in activities:

        # getting the old output relation of duplicated activity
        cursor.execute("SELECT relation_id, relation_name FROM relation WHERE relation_id = ?", (activity[2],))
        old_output_relation = cursor.fetchone()

        # # duplicate the old output relation
        # cursor.execute('''INSERT INTO relation (relation_name)
        #             VALUES (?)''', (old_output_relation[1],))

        # conn.commit()

        # # id of the new just added relation
        # output_relation_id = cursor.lastrowid

        boxType = activity[0].split("-")[0]

        # # adding a attributeRelation to each output type that this activity supports
        # for outputType in outputTypesSupported[boxType]:
        #     cursor.execute('''INSERT INTO attributeRelation (attribute_id, relation_id)
        #         VALUES (?, ?)''', (attributeIds[outputType], output_relation_id,))

        # duplicated_output_relations[old_output_relation[0]] = output_relation_id # mapping old to new ids

        # cursor.execute('''INSERT INTO activity (workflow_id, activity_name, input_relation_id, output_relation_id, ve_id)
        #         VALUES (?, ?, ?, ?, ?)''', (workflow_id, activity[0], activity[1], output_relation_id, activity[3],))

        cursor.execute('''INSERT INTO activity (workflow_id, activity_name, input_relation_id, output_relation_id, ve_id)
                VALUES (?, ?, ?, ?, ?)''', (workflow_id, activity[0], activity[1], old_output_relation[0], activity[3],))

        conn.commit()

        # id of the new just added activity
        activity_id = cursor.lastrowid

        duplicated_activities_ids.append(activity_id)

    # updating duplicated activities to point to the duplicated relations
    for old_output_id in duplicated_output_relations:
        cursor.execute("SELECT activity_id FROM activity WHERE input_relation_id = ?", (old_output_id,))
        activities = cursor.fetchall()

        for activity in activities:
            if activity[0] in duplicated_activities_ids: # this is a duplicated activity that needs to have input field updated to point to new duplicated relation
                cursor.execute("UPDATE activity SET input_relation_id = ? WHERE activity_id = ?", (duplicated_output_relations[old_output_id], activity[0],))
                conn.commit()

    # get the source activity of the connection
    cursor.execute("SELECT output_relation_id FROM activity WHERE workflow_id = ? AND activity_name = ?", (workflow_id, data['sourceNodeType']+"-"+data['sourceNodeId'],))
    source_activity = cursor.fetchone()

    # update input relation of the activity
    cursor.execute("UPDATE activity SET input_relation_id = ? WHERE workflow_id = ? AND activity_name = ?", (source_activity[0], workflow_id, data['targetNodeType']+"-"+data['targetNodeId'],))

    conn.commit()
    conn.close()

    return "",200

@bp.route('/deleteConnectionProv', methods=['POST'])
def delete_connection_prov():

    # db_path = os.path.join(os.getcwd(), ".curio", "provenance.db")
    db_path = get_db_path()
    conn = sqlite3.connect(db_path)
    cursor = conn.cursor()

    data = request.json.get('data')

    # // new version (increment version number based on previous old workflow that points to a ve that points to the version)
    # // new versioned element (pointing to the versioned element of the old workflow and pointing to the new version)
    # // new workflow
    # // point new workflow to the new versioned element
    # // duplicate all activities that point to the old workflow and point to the new one (duplicate relations tied to activities)
    # // update input relation of the activity to NULL

    # last workflow created with this name
    cursor.execute("SELECT * FROM workflow WHERE workflow_id = (SELECT MAX(workflow_id) FROM workflow WHERE workflow_name = ?)", (data['workflow_name'],))
    old_workflow = cursor.fetchone()

    # getting versionedElement attached to the old workflow
    cursor.execute("SELECT * FROM versionedElement WHERE ve_id = ?", (old_workflow[2],))
    old_workflow_ve = cursor.fetchone()

    # getting the version atteched to the old workflow
    cursor.execute("SELECT * FROM version WHERE version_id = ?", (old_workflow_ve[1],))
    version = cursor.fetchone()

    new_version_number = str(float(version[1])+1.0)

    # creating new version
    cursor.execute('''INSERT INTO version (version_number)
                    VALUES (?)''', (new_version_number,))

    conn.commit()

    # id of the new just added version
    version_id = cursor.lastrowid

    # creating new versioned element for the new workflow
    cursor.execute('''INSERT INTO versionedElement (previous_ve_id, version_id)
                    VALUES (?, ?)''', (old_workflow_ve[0], version_id,))

    conn.commit()

    # id of the new just added versioned element
    ve_id = cursor.lastrowid

    # creating new workflow
    cursor.execute('''INSERT INTO workflow (workflow_name, ve_id)
                    VALUES (?, ?)''', (data['workflow_name'], ve_id,))

    conn.commit()

    # id of the new just added workflow
    workflow_id = cursor.lastrowid

    # getting all activities that point to the old workflow
    cursor.execute("SELECT activity_name, input_relation_id, output_relation_id, ve_id FROM activity WHERE workflow_id = ?", (old_workflow[0],))

    activities = cursor.fetchall()

    duplicated_output_relations = {} # dict of old to new ids.

    duplicated_activities_ids = []

    # duplicating all activities and making them point to the new workflow
    for activity in activities:

        # getting the old output relation of duplicated activity
        cursor.execute("SELECT relation_id, relation_name FROM relation WHERE relation_id = ?", (activity[2],))
        old_output_relation = cursor.fetchone()

        # # duplicate the old output relation
        # cursor.execute('''INSERT INTO relation (relation_name)
        #             VALUES (?)''', (old_output_relation[1],))

        # conn.commit()

        # # id of the new just added relation
        # output_relation_id = cursor.lastrowid

        boxType = activity[0].split("-")[0]

        # # adding a attributeRelation to each output type that this activity supports
        # for outputType in outputTypesSupported[boxType]:
        #     cursor.execute('''INSERT INTO attributeRelation (attribute_id, relation_id)
        #         VALUES (?, ?)''', (attributeIds[outputType], output_relation_id,))

        # duplicated_output_relations[old_output_relation[0]] = output_relation_id # mapping old to new ids

        # cursor.execute('''INSERT INTO activity (workflow_id, activity_name, input_relation_id, output_relation_id, ve_id)
        #         VALUES (?, ?, ?, ?, ?)''', (workflow_id, activity[0], activity[1], output_relation_id, activity[3],))

        cursor.execute('''INSERT INTO activity (workflow_id, activity_name, input_relation_id, output_relation_id, ve_id)
                VALUES (?, ?, ?, ?, ?)''', (workflow_id, activity[0], activity[1], old_output_relation[0], activity[3],))

        conn.commit()

        # id of the new just added activity
        activity_id = cursor.lastrowid

        duplicated_activities_ids.append(activity_id)

    # updating duplicated activities to point to the duplicated relations
    for old_output_id in duplicated_output_relations:
        cursor.execute("SELECT activity_id FROM activity WHERE input_relation_id = ?", (old_output_id,))
        activities = cursor.fetchall()

        for activity in activities:
            if activity[0] in duplicated_activities_ids: # this is a duplicated activity that needs to have input field updated to point to new duplicated relation
                cursor.execute("UPDATE activity SET input_relation_id = ? WHERE activity_id = ?", (duplicated_output_relations[old_output_id], activity[0],))
                conn.commit()

    # update input relation of the activity
    cursor.execute("UPDATE activity SET input_relation_id = NULL WHERE workflow_id = ? AND activity_name = ?", (workflow_id, data['targetNodeType']+"-"+data['targetNodeId'],))

    conn.commit()
    conn.close()

    return "",200

@bp.route('/checkDB', methods=['GET'])
def check_db():
    
    # db_path = os.path.join(os.getcwd(), ".curio", "provenance.db")
    db_path = get_db_path()
    conn = sqlite3.connect(db_path)
    cursor = conn.cursor()
    return "OK",200

@bp.route('/boxExecProv', methods=['POST'])
def box_exec_prov():

    # db_path = os.path.join(os.getcwd(), ".curio", "provenance.db")
    db_path = get_db_path()
    conn = sqlite3.connect(db_path)
    cursor = conn.cursor()

    data = request.json.get('data')

    # new workflow execution
    # replicate all activity execution from old workflow execution (except the one related to the activity I'm currently running) and make them point to the new workflow execution
    # create relation instances from the input and output relation id of the activity
    # create the attribute values based on the attributes connected to the relations of the activity. The values are 1 if there is data of that type or 0 if there is not.
    # new activity execution

    # getting the id of the most recent execution of this workflow
    cursor.execute('''
        SELECT MAX(workflowexec_id)
        FROM workflowExecution
        JOIN workflow ON workflowExecution.workflow_id = workflow.workflow_id
        WHERE workflow_name = ?
    ''', (data['workflow_name'],))

    workflowexec_id = cursor.fetchone()

    # getting the most recent execution of this workflow
    cursor.execute("SELECT workflowexec_id, workflow_id, workflowexec_start_time, workflowexec_end_time FROM workflowExecution WHERE workflowexec_id = ?", (workflowexec_id[0],))
    old_workflow_execution = cursor.fetchone()

    # getting last workflow create with this name
    cursor.execute("SELECT * FROM workflow WHERE workflow_id = (SELECT MAX(workflow_id) FROM workflow WHERE workflow_name = ?)", (data['workflow_name'],))
    workflow = cursor.fetchone()

    # getting activity attached to this workflow
    cursor.execute("SELECT activity_id, input_relation_id, output_relation_id FROM activity WHERE workflow_id = ? AND activity_name = ?", (workflow[0], data['activity_name'],))
    activity = cursor.fetchone()

    # creating new workflow execution
    cursor.execute('''INSERT INTO workflowExecution (workflowexec_start_time, workflowexec_end_time, workflow_id)
                    VALUES (?, ?, ?)''', (data["activityexec_start_time"], data["activityexec_end_time"], workflow[0],))

    conn.commit()

    # id of the new just added workflowExecution
    workflowExecution_id = cursor.lastrowid

    old_activity_executions = []

    if(old_workflow_execution != None):
        # getting all activity execution from the old workflow execution except the one related to the activity
        cursor.execute("SELECT activity_id, activityexec_start_time, activityexec_end_time, input_ri_id, output_ri_id, activity_source_code FROM activityExecution WHERE workflowexec_id = ? AND activity_id <> ?", (old_workflow_execution[0], activity[0],))
        old_activity_executions = cursor.fetchall()

    input_relation_id = activity[1]
    output_relation_id = activity[2]

    # get the most recent relationInstance of the input relation
    cursor.execute("SELECT re_id FROM relationInstance WHERE re_id = (SELECT MAX(re_id) FROM relationInstance WHERE relation_id = ?)", (input_relation_id,))
    relation_instance_input = cursor.fetchone()

    if(relation_instance_input != None):
        # creating relation instances
        cursor.execute('''INSERT INTO relationInstance (relation_id, original_ri)
                        VALUES (?, ?)''', (input_relation_id, relation_instance_input[0],))
    else:
        # creating relation instances
        cursor.execute('''INSERT INTO relationInstance (relation_id)
                        VALUES (?)''', (input_relation_id,))

    conn.commit()

    input_ri_id = cursor.lastrowid

    # # get the most recent relationInstance of the output relation
    # cursor.execute("SELECT re_id FROM relationInstance WHERE re_id = (SELECT MAX(re_id) FROM relationInstance WHERE relation_id = ?)", (output_relation_id,))
    # relation_instance_output = cursor.fetchone()

    # if(relation_instance_output != None):
    #     cursor.execute('''INSERT INTO relationInstance (relation_id, original_ri)
    #                     VALUES (?)''', (output_relation_id, relation_instance_output[0]))
    # else:
    #     # creating relation instances
    #     cursor.execute('''INSERT INTO relationInstance (relation_id)
    #                     VALUES (?)''', (output_relation_id,))

    # creating relation instances
    cursor.execute('''INSERT INTO relationInstance (relation_id)
                    VALUES (?)''', (output_relation_id,))

    conn.commit()

    output_ri_id = cursor.lastrowid

    # getting types names and ids of attributes of input relation
    cursor.execute('''SELECT attribute.attribute_id, attribute.attribute_name
                        FROM relation
                        JOIN attributeRelation ON relation.relation_id = attributeRelation.relation_id
                        JOIN attribute ON attributeRelation.attribute_id = attribute.attribute_id
                        WHERE attribute.attribute_type = ? AND relation.relation_id = ?''', ('Data', input_relation_id))

    input_attributes = cursor.fetchall()

    # getting types names and ids of attributes of output relation
    cursor.execute('''SELECT attribute.attribute_id, attribute.attribute_name
                        FROM relation
                        JOIN attributeRelation ON relation.relation_id = attributeRelation.relation_id
                        JOIN attribute ON attributeRelation.attribute_id = attribute.attribute_id
                        WHERE attribute.attribute_type = ? AND relation.relation_id = ?''', ('Data', output_relation_id))

    output_attributes = cursor.fetchall()
    # print(output_attributes)

    for input_attribute in input_attributes:
        # creating attributeValues
        cursor.execute('''INSERT INTO attributeValue (attribute_id, ri_id, value)
                        VALUES (?, ?, ?)''', (input_attribute[0], input_ri_id, data['types_input'][input_attribute[1]],))

        conn.commit()

    for output_attribute in output_attributes:
        # creating attributeValues
        cursor.execute('''INSERT INTO attributeValue (attribute_id, ri_id, value)
                        VALUES (?, ?, ?)''', (output_attribute[0], output_ri_id, data['types_output'][output_attribute[1]],))

        conn.commit()

    # duplicating all old activity executions of this workflow
    for old_activity_execution in old_activity_executions:
        cursor.execute('''INSERT INTO activityExecution (activity_id, workflowexec_id, activityexec_start_time, activityexec_end_time, input_ri_id, output_ri_id, activity_source_code)
                VALUES (?, ?, ?, ?, ?, ?, ?)''', (old_activity_execution[0], workflowExecution_id, old_activity_execution[1], old_activity_execution[2], old_activity_execution[3], old_activity_execution[4], old_activity_execution[5]))

        conn.commit()

    # adding new activityExecution
    cursor.execute('''INSERT INTO activityExecution (activity_id, workflowexec_id, activityexec_start_time, activityexec_end_time, input_ri_id, output_ri_id, activity_source_code)
            VALUES (?, ?, ?, ?, ?, ?, ?)''', (activity[0], workflowExecution_id, data['activityexec_start_time'], data['activityexec_end_time'], input_ri_id, output_ri_id, data["activity_source_code"]))

    conn.commit()
    conn.close()

    return "",200

@bp.route('/getBoxGraph', methods=['POST'])
def get_box_graph():

    # db_path = os.path.join(os.getcwd(), ".curio", "provenance.db")
    db_path = get_db_path()
    conn = sqlite3.connect(db_path)
    cursor = conn.cursor()

    data = request.json.get('data')

    # last workflow created with this name
    cursor.execute("SELECT workflow_id FROM workflow WHERE workflow_id = (SELECT MAX(workflow_id) FROM workflow WHERE workflow_name = ?)", (data['workflow_name'],))
    workflow = cursor.fetchone()

    # activity
    cursor.execute("SELECT activity_id FROM activity WHERE workflow_id = ? AND activity_name = ?", (workflow[0], data['activity_name'],))
    activity = cursor.fetchone()

    # all the executions of this activity in the order they were executed
    cursor.execute("SELECT activityexec_id, input_ri_id, output_ri_id, activity_source_code FROM activityExecution WHERE activity_id = ? ORDER BY activityexec_id ASC", (activity[0],))
    activity_executions = cursor.fetchall()

    # getting the list of all attribute values for input of all executions of this activity
    cursor.execute('''SELECT activityExecution.activityexec_id, attribute.attribute_name
                        FROM activityExecution
                        JOIN attributeValue ON activityExecution.input_ri_id = attributeValue.ri_id
                        JOIN attribute ON attribute.attribute_id = attributeValue.attribute_id
                        WHERE activityExecution.activity_id = ? AND value = ? AND attribute.attribute_type = ?''', (activity[0], '1', 'Data'))
    input_types = cursor.fetchall()

    # getting the list of all attribute values for input of all executions of this activity
    cursor.execute('''SELECT activityExecution.activityexec_id, attribute.attribute_name
                        FROM activityExecution
                        JOIN attributeValue ON activityExecution.output_ri_id = attributeValue.ri_id
                        JOIN attribute ON attribute.attribute_id = attributeValue.attribute_id
                        WHERE activityExecution.activity_id = ? AND value = ? AND attribute.attribute_type = ?''', (activity[0], '1', 'Data'))
    output_types = cursor.fetchall()

    mapIdToTypes_input = {} # activity execution id to array of input types
    mapIdToTypes_output = {} # activity execution id to array of output types

    for input_type in input_types:

        if input_type[0] not in mapIdToTypes_input:
            mapIdToTypes_input[input_type[0]] = [input_type[1]]
        else:
            mapIdToTypes_input[input_type[0]].append(input_type[1])

    for output_type in output_types:

        if output_type[0] not in mapIdToTypes_output:
            mapIdToTypes_output[output_type[0]] = [output_type[1]]
        else:
            mapIdToTypes_output[output_type[0]].append(output_type[1])

    graph = []

    for activity_execution in activity_executions:

        if activity_execution[0] not in mapIdToTypes_input:
            mapIdToTypes_input[activity_execution[0]] = []

        if activity_execution[0] not in mapIdToTypes_output:
            mapIdToTypes_output[activity_execution[0]] = []

        node = {
            "id": activity_execution[0],
            "code": activity_execution[3],
            "inputs": mapIdToTypes_input[activity_execution[0]],
            "outputs": mapIdToTypes_output[activity_execution[0]]
        }

        graph.append(node)

    conn.commit()
    conn.close()

    return jsonify({
        "graph": graph
    }),200

@bp.route('/truncateDBProv', methods=['GET'])
def truncate_db_prov():

    # db_path = os.path.join(os.getcwd(), ".curio", "provenance.db")
    db_path = get_db_path()
    conn = sqlite3.connect(db_path)
    cursor = conn.cursor()

    # Fetch the list of tables in the database
    cursor.execute("SELECT name FROM sqlite_master WHERE type='table';")
    tables = cursor.fetchall()
    for table in tables:
        # Get the table name from the tuple
        table_name = table[0]
        # Execute a DELETE statement to truncate the table
        cursor.execute(f"DELETE FROM {table_name};")

    cursor.execute("INSERT INTO ATTRIBUTE(attribute_id, attribute_name, attribute_type) VALUES (1, 'DATAFRAME', 'Data');")
    cursor.execute("INSERT INTO ATTRIBUTE(attribute_id, attribute_name, attribute_type) VALUES (2, 'GEODATAFRAME', 'Data');")
    cursor.execute("INSERT INTO ATTRIBUTE(attribute_id, attribute_name, attribute_type) VALUES (3, 'VALUE', 'Data');")
    cursor.execute("INSERT INTO ATTRIBUTE(attribute_id, attribute_name, attribute_type) VALUES (4, 'LIST', 'Data');")
    cursor.execute("INSERT INTO ATTRIBUTE(attribute_id, attribute_name, attribute_type) VALUES (5, 'JSON', 'Data');")

    conn.commit()
    conn.close()

    return "",200



def create_template_object(folder, filename, code):
    return {
        "id": str(uuid.uuid4()),
        "type": TYPE_MAP.get(folder, "UNKNOWN"),
        "name": filename.replace(".py", "").replace("_", " "),
        "description": "",
        "accessLevel": "ANY",
        "code": code,
        "custom": True
    }

def generate_templates():
    templates = []

    for folder in TYPE_MAP.keys():
        folder_path = os.path.join(TEMPLATE_DIR, folder)

        if not os.path.isdir(folder_path):
            continue

        for file in os.listdir(folder_path):
            if file.endswith(".py"):

                with open(os.path.join(folder_path, file), "r", encoding="utf-8") as f:
                    code = f.read()

                template_obj = create_template_object(folder, file, code)
                templates.append(template_obj)

    return templates

@bp.route('/datasets', methods=['GET'])
def list_datasets():
    response = requests.get(api_address+":"+str(api_port)+"/datasets")
    response.raise_for_status() 
    files = response.json()
    return jsonify(files)
    

@bp.route("/templates", methods=["GET"])
def get_templates():
    return jsonify(generate_templates())

@bp.route('/addTemplate', methods=['POST'])
def add_template():
    data = request.get_json()

    required_fields = ['id', 'type', 'name', 'description', 'accessLevel', 'code', 'custom']
    if not all(field in data for field in required_fields):
        return jsonify({'error': 'Missing one or more required fields'}), 400

    template_type = data['type']
    if template_type not in FOLDER_MAP:
        return jsonify({'error': f"Unknown template type: {template_type}"}), 400

    subfolder = FOLDER_MAP[template_type]
    folder_path = os.path.join(TEMPLATE_DIR, subfolder)
    os.makedirs(folder_path, exist_ok=True)

    replace_name = data['name'].replace(" ", "_")

    filename = f"{replace_name}.py"
    filepath = os.path.join(folder_path, filename)

    try:
        with open(filepath, 'w', encoding='utf-8') as f:
            f.write(data['code'])
    except Exception as e:
        return jsonify({'error': str(e)}), 500

    return jsonify({'message': f"Template saved to {filepath}"}), 200



{
  "$schema": "https://vega.github.io/schema/vega-lite/v5.json",
  "description": "A simple bar chart with embedded data.",
  "mark": "bar",
  "encoding": {
    "x": {"field": "a", "type": "nominal", "axis": {"labelAngle": 0}},
    "y": {"field": "b", "type": "quantitative"}
  }
}


from utk_curio.sandbox.app import app
import os

@app.route('/health', methods=['GET'])
def health():
    return 'OK', 200

if __name__ == '__main__':
    app.run(
        host=os.getenv('FLASK_SANDBOX_HOST', 'localhost'),
        port=int(os.getenv('FLASK_SANDBOX_PORT', 2000)),
        threaded=False,
        debug=True
    )




from flask import request, abort, jsonify
import json
import subprocess
import geopandas as gpd
import pandas as pd
import utk
from utk_curio.sandbox.app import app, cache
from utk_curio.sandbox.app.utils.cache import make_key
import os
import mmap
from pathlib import Path

from shapely import wkt

DATA_DIR = "./data"

@app.after_request
def add_cors_headers(response):
    response.headers.add('Access-Control-Allow-Origin', '*')
    response.headers.add('Access-Control-Allow-Headers', 'Content-Type,Authorization')
    response.headers.add('Access-Control-Allow-Methods', 'GET,PUT,POST,DELETE')
    return response

@app.route('/')
def root():
    abort(403)

@app.route('/live', methods=['GET'])
def live():
    return 'Sandbox is live.'

@app.route('/upload', methods=['POST'])
def upload_file():
    if 'file' not in request.files:
        return 'No file part'

    file = request.files['file']
    if file.filename == '':
        return 'No selected file'

    file.save(request.form['fileName'])

    return file.filename

@app.route('/datasets', methods=['GET'])
def list_datasets():
    allowed_extensions = {'.json', '.geojson', '.csv'}

    files = []

    # Source 1: /data relative to the root of the installed pip package
    project_root_data = Path(__file__).parent.parent.parent / 'data'
    print("Loading datasets from pip package location:", project_root_data)

    if project_root_data.exists() and project_root_data.is_dir():
        files.extend([
            f.as_posix() for f in project_root_data.iterdir()
            if f.is_file() and f.suffix.lower() in allowed_extensions
        ])

    # Source 2: /data relative to current working directory
    # cwd_data = os.getcwd() / 'data'
    launch_dir = os.environ.get("CURIO_LAUNCH_CWD", os.getcwd())
    data_dir = os.path.join(launch_dir, "data")
    data_dir = Path(data_dir)
    print("Loading datasets from working directory:", data_dir)

    if data_dir.exists() and data_dir.is_dir():
        files.extend([
            f.as_posix() for f in data_dir.iterdir()
            if f.is_file() and f.suffix.lower() in allowed_extensions
        ])

    return jsonify(files)

@app.route('/exec', methods=['POST'])
# @cache.cached(make_cache_key=make_key)
def exec():
    import time
    start_time = time.time()
    app.logger.info(f'/exec: Request begin')

    # print(request.json['code'], flush=True)

    if(request.json['code'] == None):
        abort(400, "Code was not included in the post request")

    # Load default python wrapper code
    full_code = open('sandbox/python_wrapper.txt', 'r').read()

    # Set path to be relative to the place where curio is called
    original_dir = os.getcwd()
    launch_dir = os.environ.get("CURIO_LAUNCH_CWD", os.getcwd())
    os.chdir(launch_dir)

    code = request.json['code']
    file_path = request.json['file_path']
    boxType = request.json['boxType']
    dataType = request.json['dataType']
    
    full_code = full_code.replace('{userCode}', str(code))
    full_code = full_code.replace('{filePath}', str(file_path))
    full_code = full_code.replace('{boxType}', str(boxType))
    full_code = full_code.replace('{dataType}', str(dataType))

    print("File input:", file_path)

    command = ['python', '-']
    process = subprocess.Popen(command, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
    stdout, stderr = process.communicate(full_code)

    stdout = [item for item in stdout.split("\n") if item != '']

    print("File output", stdout)

    if(len(stdout) > 0):
        output = json.loads(stdout[-1])
    else:
        output = {}
        output['path'] = ""
        output['dataType'] = "str"

    jsonOutput = {
        "stdout": stdout[0:-1], # just get prints, remove output itself
        "stderr": stderr,
        "output": output
    }

    # print("----------", jsonOutput, flush=True)

    app.logger.info(f'/exec: Request end in time: {(time.time() - start_time) / 60} mins')

    os.chdir(original_dir)

    return jsonify(jsonOutput)

@app.route('/toLayers', methods=['POST'])
def toLayers():

    if(request.json['geojsons'] == None):
        abort(400, "geojsons were not included in the post request")

    geojsons = request.json['geojsons']

    layers = []
    joinedJsons = []

    for index, geojson in enumerate(geojsons):

        parsedGeoJson = geojson # json.loads(geojson)

        layerName = "layer"+str(index)

        if 'metadata' in parsedGeoJson and 'name' in parsedGeoJson['metadata']:
            layerName = parsedGeoJson['metadata']['name']

        # gdfs.append(gpd.GeoDataFrame.from_features(geoJson))
        gdf = gpd.GeoDataFrame.from_features(parsedGeoJson)
        # df = pd.DataFrame.from_dict(geojson)
        # df = pd.DataFrame({'geometry': geojson['geometry'], 'values': geojson['value']})
        # df = df[df['geometry'].apply(lambda x: isinstance(x, str))]
        # df['geometry'] = df['geometry'].apply(wkt.loads)
        # gdf = gpd.GeoDataFrame(df, geometry='geometry')

        if 'building_id' in gdf.columns:

            gdf = gdf.set_crs('4326')
            mesh = utk.OSM.mesh_from_buildings_gdf(gdf, 5)['data']

            non_geometry_columns = [col for col in gdf.columns if col != gdf.geometry.name and col != "id" and col != "interacted" and col != "linked" and col != 'building_id' and col != 'tags' and col != 'height' and col != 'min_height']

            joinedJson = {
                "id": layerName,
                "incomingId": [],
                "inValues": []
            }

            renderStyle = []

            if(len(non_geometry_columns) > 0):
                renderStyle = ["SMOOTH_COLOR_MAP_TEX", "PICKING"]
            else:
                renderStyle = ["SMOOTH_COLOR_MAP_TEX"]

            layer = {
                "id": layerName,
                "type": "BUILDINGS_LAYER",
                "renderStyle": renderStyle,
                "styleKey": "surface",
                "data": mesh
            }

            layers.append(layer)

            for column in non_geometry_columns:

                inValues = []

                currentBuildingId = -1

                uniqueObjectIndex = 0

                print("column", column)

                for index, row in gdf.iterrows():

                    if(row['building_id'] != currentBuildingId): # only replicate values for the first reference to that building
                        currentBuildingId = row['building_id']

                        objectUnit = layer['data'][uniqueObjectIndex]['geometry'] # object (each row of the gdf was transformed in a set of coordinates)

                        for i in range(int(len(objectUnit['coordinates'])/3)):
                            if(isinstance(row[column],list)): # different values for each coordinate # TODO: consider multiple timesteps
                                inValues.append(row[column][i])
                            else: # for each coordinate replicate the value of the row
                                inValues.append(row[column])

                        uniqueObjectIndex += 1

                joinedJson["incomingId"].append(column)
                joinedJson["inValues"].append([inValues]) # TODO: support for multiple timesteps

            joinedJsons.append(joinedJson)

        elif 'surface_id' in gdf.columns:

            gdf = gdf.set_crs('3395')
            gdf = gdf.to_crs('4326')

            polygon_geometry = gdf.geometry.iloc[0]

            coordinates = list(polygon_geometry.exterior.coords)

            minLat = None
            maxLat = None
            minLon = None
            maxLon = None

            for coord in coordinates:
                if(minLat == None or minLat > coord[1]):
                    minLat = coord[1]

                if(maxLat == None or maxLat < coord[1]):
                    maxLat = coord[1]

                if(minLon == None or minLon > coord[0]):
                    minLon = coord[0]

                if(maxLon == None or maxLon < coord[0]):
                    maxLon = coord[0]

            mesh = utk.OSM.create_surface_mesh([minLat, minLon, maxLat, maxLon], True, -1, 5)

            non_geometry_columns = [col for col in gdf.columns if col != gdf.geometry.name and col != "id" and col != "interacted" and col != "linked" and col != 'surface_id']

            joinedJson = {
                "id": layerName,
                "incomingId": [],
                "inValues": []
            }

            renderStyle = []

            if(len(non_geometry_columns) > 0):
                renderStyle = ["SMOOTH_COLOR_MAP", "PICKING"]
            else:
                renderStyle = ["SMOOTH_COLOR"]

            layer = {
                "id": layerName,
                "type": "TRIANGLES_3D_LAYER",
                "renderStyle": renderStyle,
                "styleKey": "surface",
                "data": mesh['data']
            }

            layers.append(layer)

            for column in non_geometry_columns:

                inValues = []

                for index, row in gdf.iterrows():

                    objectUnit = layer['data'][index]['geometry'] # object (each row of the gdf was transformed in a set of coordinates)

                    for i in range(int(len(objectUnit['coordinates'])/3)):
                        if(isinstance(row[column],list)): # different values for each coordinate # TODO: consider multiple timesteps
                            inValues.append(row[column][i])
                        else: # for each coordinate replicate the value of the row
                            inValues.append(row[column])

                joinedJson["incomingId"].append(column)
                joinedJson["inValues"].append([inValues]) # TODO: support for multiple timesteps

            joinedJsons.append(joinedJson)

        else:

            gdf = gdf.set_crs('3395')
            mesh = utk.mesh_from_gdf(gdf)

            # layer = {
            #     "id": layerName,
            #     "type": "TRIANGLES_3D_LAYER",
            #     "renderStyle": ["SMOOTH_COLOR_MAP"],
            #     "styleKey": "surface",
            #     "data": mesh
            # }

            non_geometry_columns = [col for col in gdf.columns if col != gdf.geometry.name and col != "id" and col != "interacted" and col != "linked"]

            joinedJson = {
                "id": layerName,
                "incomingId": [],
                "inValues": []
            }

            renderStyle = []

            if(len(non_geometry_columns) > 0):
                renderStyle = ["SMOOTH_COLOR_MAP", "PICKING"]
            else:
                renderStyle = ["SMOOTH_COLOR"]

            layer = {
                "id": layerName,
                "type": "TRIANGLES_3D_LAYER",
                "renderStyle": renderStyle,
                "styleKey": "surface",
                "data": mesh
            }

            layers.append(layer)

            for column in non_geometry_columns:

                inValues = []

                for index, row in gdf.iterrows():
                    # print(layer['data'])
                    # print(layer['data'], flush=True)

                    objectUnit = layer['data'][index]['geometry'] # object (each row of the gdf was transformed in a set of coordinates)
                    
                    for i in range(int(len(objectUnit['coordinates'])/3)):
                        if(isinstance(row[column],list)): # different values for each coordinate # TODO: consider multiple timesteps
                            inValues.append(row[column][i])
                        else: # for each coordinate replicate the value of the row
                            inValues.append(row[column])

                joinedJson["incomingId"].append(column)
                joinedJson["inValues"].append([inValues]) # TODO: support for multiple timesteps

            joinedJsons.append(joinedJson)

    jsonOutput = {
        "layers": layers,
        "joinedJsons": joinedJsons
    }

    return jsonify(jsonOutput)




# from shadowAccumulator import *
from .load_utk import *
from .utils import *
from .load_physical import *
from .load_thematic import *

try:
    # Avoid importing it in systems without optix
    from .shadow_accumulator import *
except:
    import warnings
    warnings.warn("Skipping shadow_accumulator import.")

# exposes simulations like .shadow
# loads any kind of data the type is determined by the extension
def shadow(data, intervals):

    coordinates = []

    for value in data:

        layer_json = {}

        if(type(value) == str): # it is a filepath
            layer_json = load_utk(value)
        else:
            layer_json = value

        coordinates += get_coordinates(layer_json)

    latitudes = []
    longitudes = []

    for i in range(int(len(coordinates)/3)):
        
        latitudes.append(coordinates[i*3])
        longitudes.append(coordinates[i*3+1])

    centroid = convert_projections('3395', '4326', [(min(longitudes) + max(longitudes))/2, (min(latitudes) + max(latitudes))/2])

    shadowAccumulator = ShadowAccumulator(centroid[1], centroid[0], data, intervals)
    shadowAccumulator.accumulate_shadow()

    return shadowAccumulator


import os
from dotenv import load_dotenv


basedir = os.path.abspath(os.path.dirname(__file__))
load_dotenv(os.path.join(basedir, '.env'))
load_dotenv(os.path.join(basedir, '.flaskenv'))

class Config:
    LOG_TO_STDOUT = os.environ.get('LOG_TO_STDOUT')

    SECRET_KEY = os.environ.get('SECRET_KEY') or 'you-will-never-guess'

    SQLALCHEMY_DATABASE_URI = os.environ.get('DATABASE_URL') or \
                              'sqlite:///urban_workflow.db'
    SQLALCHEMY_TRACK_MODIFICATIONS = True

    WTF_CSRF_ENABLED = True

    ADMINS = ['your-email@example.com']

    SANDBOX_RELATIVE_PATH = (os.environ.get('SANDBOX_RELATIVE_PATH') or
                             '../sandbox')
    FRONTEND_RELATIVE_PATH = (os.environ.get('FRONTEND_RELATIVE_PATH') or
                                '../urban-workflows')
    FRONTEND_PORT = os.environ.get('FRONTEND_PORT') or 3000
    SANDBOX_PORT = os.environ.get('SANDBOX_PORT') or 2000




from bin.consts import fields, fakeScatterData, fakeHmat, fakeScatterSys
import random

class Structure(object):
    def __init__(self) -> None:
        pass

    def getHmat(self, pt, time):
        return fakeHmat
    
    def getScatter(self, fld1, fld2):
        msg = f"[Structure - getScatter] {fld1} {fld2}"

        print(msg)
        # return fakeScatterData
        return fakeScatterSys
    
    def getTemporalData(self, fieldKeys):
        temporalData = {}

        for f in fieldKeys:
            temporalData[f] = []

            for i in range(120):
                v = random.randint(0, 100)

                if i < 96: 
                    t = i + 1 
                    c = "WRFout"
                else:
                    t = i - 95
                    c = "Obs"
                
                obj = {"t": t, "v": v, "c": c, "f": "T" }
                temporalData[f].append(obj)

        return temporalData

    def getFields(self):
        return fields


import pandas as pd

df = arg # Getting DataFrame from a previous node

df.dropna(how='all', inplace=True)

df['Age'] = pd.to_numeric(df['Age'], errors='coerce')  # convert to numeric
df['Age'].fillna(df['Age'].mean(), inplace=True)

df['Name'].fillna('Unknown', inplace=True)
df['Email'].fillna('no-email@example.com', inplace=True)

df['Name'] = df['Name'].str.strip().str.title()  # ' alice ' -> 'Alice'

df.drop_duplicates(inplace=True)

return df


from typing import Union, Tuple

# from numba import jit
'''
 The MIT License (MIT)
 Copyright (c) 2014 Michael Hirsch
 reference: http://en.wikipedia.org/wiki/Cohen%E2%80%93Sutherland_algorithm
 * The best way to Numba JIT this would probably be in the function calling this,
   to include the loop itself inside the jit decoration.
'''

# @jit

INSIDE, LEFT, RIGHT, LOWER, UPPER = 0, 1, 2, 4, 8
def cohensutherland(
    xmin: float, ymax: float, xmax: float, ymin: float, x1: float, y1: float, x2: float, y2: float
) -> Tuple[float, float, float, float]:
    """Clips a line to a rectangular area.
    This implements the Cohen-Sutherland line clipping algorithm.  xmin,
    ymax, xmax and ymin denote the clipping area, into which the line
    defined by x1, y1 (start point) and x2, y2 (end point) will be
    clipped.
    If the line does not intersect with the rectangular clipping area,
    four None values will be returned as tuple. Otherwise a tuple of the
    clipped line points will be returned in the form (cx1, cy1, cx2, cy2).
    """

    def _getclip(xa, ya):
        # if dbglvl>1: print('point: '),; print(xa,ya)
        p = INSIDE  # default is inside

        # consider x
        if xa <= xmin:
            p |= LEFT
        elif xa >= xmax:
            p |= RIGHT

        # consider y
        if ya <= ymin:
            p |= LOWER  # bitwise OR
        elif ya >= ymax:
            p |= UPPER  # bitwise OR
        return p

    # check for trivially outside lines
    k1 = _getclip(x1, y1)
    k2 = _getclip(x2, y2)

    # %% examine non-trivially outside points
    # bitwise OR |
    opt = -1
    while (k1 | k2) != 0:  # if both points are inside box (0000) , ACCEPT trivial whole line in box

        # if line trivially outside window, REJECT
        if (k1 & k2) != 0:  # bitwise AND &
            # if dbglvl>1: print('  REJECT trivially outside box')
            # return nan, nan, nan, nan
            return None

        # non-trivial case, at least one point outside window
        # this is not a bitwise or, it's the word "or"
        opt = k1 or k2  # take first non-zero point, short circuit logic
        if opt & UPPER:  # these are bitwise ANDS
            x = x1 + (x2 - x1) * (ymax - y1) / (y2 - y1)
            y = ymax
        elif opt & LOWER:
            x = x1 + (x2 - x1) * (ymin - y1) / (y2 - y1)
            y = ymin
        elif opt & RIGHT:
            y = y1 + (y2 - y1) * (xmax - x1) / (x2 - x1)
            x = xmax
        elif opt & LEFT:
            y = y1 + (y2 - y1) * (xmin - x1) / (x2 - x1)
            x = xmin
        else:
            raise RuntimeError('Undefined clipping state')

        if opt == k1:
            x1, y1 = x, y
            k1 = _getclip(x1, y1)
            # if dbglvl>1: print('checking k1: ' + str(x) + ',' + str(y) + '    ' + str(k1))
        elif opt == k2:
            # if dbglvl>1: print('checking k2: ' + str(x) + ',' + str(y) + '    ' + str(k2))
            x2, y2 = x, y
            k2 = _getclip(x2, y2)
    print(k1,k2)
    return x1, y1, x2, y2, opt


#!/usr/bin/env python3

import subprocess
import os
import sys
import time
import threading
import queue
import argparse
import signal
import platform

from http.server import ThreadingHTTPServer, SimpleHTTPRequestHandler

output_queue = queue.Queue()
shell_required = platform.system() == "Windows"

# Ensure unbuffered output (immediate print)
os.environ["PYTHONUNBUFFERED"] = "1"
sys.stdout.reconfigure(line_buffering=True)
sys.stderr.reconfigure(line_buffering=True)

# ANSI color codes for clear distinction
COLOR_RESET = "\033[0m"
COLOR_FRONTEND = "\033[96m"  # Cyan
COLOR_BACKEND = "\033[92m"   # Green
COLOR_SANDBOX = "\033[93m"   # Yellow

shutdown_flag = threading.Event()
processes = []

def stream_output(process, name, color):
    """Safely stream subprocess output."""
    try:
        while process.poll() is None and not shutdown_flag.is_set():
            # Read output line by line
            line = process.stdout.readline()
            if line:
                print(f"{color}[{name}] {line.strip()}{COLOR_RESET}")
        
        print(f"{color}[{name}] has stopped. No more output.{COLOR_RESET}")
    except ValueError:
        print(f"{color}[{name}] Error: Output stream closed unexpectedly.{COLOR_RESET}")
    finally:
        # Ensure the output streams are properly closed
        if process.stdout:
            process.stdout.close()
        if process.stderr:
            process.stderr.close()

def set_environment_variables(backend_host, backend_port, sandbox_host, sandbox_port):
    """Sets the environment variables for Backend and Sandbox."""
    os.environ["FLASK_BACKEND_HOST"] = backend_host
    os.environ["FLASK_BACKEND_PORT"] = str(backend_port)
    os.environ["FLASK_SANDBOX_HOST"] = sandbox_host
    os.environ["FLASK_SANDBOX_PORT"] = str(sandbox_port)
    os.environ["CURIO_LAUNCH_CWD"] = os.getcwd()
    
    print(f"Environment Variables Set:")
    print(f"FLASK_BACKEND_HOST={os.environ['FLASK_BACKEND_HOST']}")
    print(f"FLASK_BACKEND_PORT={os.environ['FLASK_BACKEND_PORT']}")
    print(f"FLASK_SANDBOX_HOST={os.environ['FLASK_SANDBOX_HOST']}")
    print(f"FLASK_SANDBOX_PORT={os.environ['FLASK_SANDBOX_PORT']}")
    print(f"CURIO_LAUNCH_CWD={os.environ['CURIO_LAUNCH_CWD']}")

def logger():
    """
    Continuously reads from the queue and prints to the terminal.
    """
    while True:
        line = output_queue.get()
        if line is None:
            break
        print(line)
        output_queue.task_done()

def check_install_build(dir, force_rebuild=False):
    # Determine the absolute path whether it is provided as relative or absolute
    script_dir = os.path.dirname(os.path.abspath(__file__))
    abs_dir = os.path.abspath(dir) if os.path.isabs(dir) else os.path.join(script_dir, dir)
    
    if not os.path.exists(abs_dir):
        raise FileNotFoundError(f"[Error] The directory '{abs_dir}' does not exist.")
    
    os.chdir(abs_dir)
    print(f"{COLOR_FRONTEND}[Frontend] Current working directory for npm commands: {os.getcwd()}{COLOR_RESET}")
    
    if force_rebuild:
        print(f"{COLOR_FRONTEND}[Frontend] Force rebuilding in {dir}...")
        if os.path.exists("node_modules"):
            subprocess.run(["rm", "-rf", "node_modules"], check=True)
        if os.path.exists("dist"):
            subprocess.run(["rm", "-rf", "dist"], check=True)
        if os.path.exists("build"):
            subprocess.run(["rm", "-rf", "build"], check=True)
    
    # Check if node_modules exist, if not, run npm install
    if not os.path.exists("node_modules"):
        print(f"{COLOR_FRONTEND}[Frontend] node_modules not found. Running npm install...{COLOR_RESET}")
        try:
            subprocess.run(["npm", "install"], check=True, shell=shell_required)
        except Exception as e:
            print(f"{COLOR_FRONTEND}[Frontend] Unexpected Error: {str(e)}{COLOR_RESET}")
            clean_shutdown()
    else:
        print(f"{COLOR_FRONTEND}[Frontend] node_modules directory already exists. Skipping npm install.{COLOR_RESET}")

    # Check if dist/build directory exists (depending on your setup)
    build_dir = "dist" if os.path.exists("dist") else "build"
    if not os.path.exists(build_dir):
        print(f"{COLOR_FRONTEND}[Frontend] {build_dir} directory not found. Running npm run build...{COLOR_RESET}")
        try:
            subprocess.run(["npm", "run", "build"], check=True, shell=shell_required)
        except Exception as e:
            print(f"{COLOR_FRONTEND}[Frontend] Unexpected Error: {str(e)}{COLOR_RESET}")
            clean_shutdown()
    else:
        print(f"{COLOR_FRONTEND}[Frontend] {build_dir} directory exists. Skipping npm run build.{COLOR_RESET}")

def force_rebuild_frontend():
    print(f"{COLOR_FRONTEND}[Frontend] Force rebuild requested.{COLOR_RESET}")
    check_install_build("frontend/urban-workflows/", force_rebuild=True)
    check_install_build("frontend/utk-workflow/src/utk-ts", force_rebuild=True)
    print(f"{COLOR_FRONTEND}[Frontend] Force rebuild complete.{COLOR_RESET}")

def start_frontend(force_rebuild=False, no_server=False):

    # Only check if running dev mode
    original_dir = os.getcwd()
    if os.getenv("CURIO_DEV") == "1":
        check_install_build("frontend/utk-workflow/src/utk-ts", force_rebuild=force_rebuild)
        os.chdir(original_dir)
        check_install_build("frontend/urban-workflows/", force_rebuild=force_rebuild)
        os.chdir(original_dir)

    # If we're not starting the server, just exit here
    if no_server:
        print(f"{COLOR_FRONTEND}[Frontend] Build completed with --force-rebuild, server not started.{COLOR_RESET}")
        return None

    dir = "frontend/urban-workflows/"
    script_dir = os.path.dirname(os.path.abspath(__file__))
    abs_dir = os.path.abspath(dir) if os.path.isabs(dir) else os.path.join(script_dir, dir)
    os.chdir(abs_dir)
    print(f"{COLOR_FRONTEND}[Frontend] Current working directory: {os.getcwd()}{COLOR_RESET}")

    try:

        if os.environ.get("CURIO_DEV") == "1":

            # Start the Node.js server
            process = subprocess.Popen(
                ["npm", "run", "start"],
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                text=True,
                shell=shell_required,
                env={**os.environ}
            )
            threading.Thread(target=stream_output, args=(process, "Frontend", COLOR_FRONTEND), daemon=True).start()

            # Check if process exited unexpectedly
            if process.poll() is not None:
                print(f"{COLOR_FRONTEND}[Frontend] Error: Server exited with code {process.returncode}{COLOR_RESET}")
                print(f"{COLOR_FRONTEND}[Frontend] Error Output:\n{process.stderr.read()}{COLOR_RESET}")
                clean_shutdown()

        else:
            process = subprocess.Popen(
                ["python", "-m", "http.server", "8080", "--directory", "dist"],
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                text=True,
                shell=shell_required,
                env={**os.environ}
            )
            print(f"[Frontend] Serving static files.")

    except subprocess.CalledProcessError as e:
        print(f"{COLOR_FRONTEND}[Frontend] Exit Code: {e.returncode}{COLOR_RESET}")
        print(f"{COLOR_FRONTEND}[Frontend] Output:\n{e.output}{COLOR_RESET}")
        print(f"{COLOR_FRONTEND}[Frontend] Error Output:\n{e.stderr}{COLOR_RESET}")
    except subprocess.TimeoutExpired:
        print(f"{COLOR_FRONTEND}[Frontend] Error: Server took too long to start.{COLOR_RESET}")
        clean_shutdown()
    except Exception as e:
        print(f"{COLOR_FRONTEND}[Frontend] Unexpected Error: {str(e)}{COLOR_RESET}")
        clean_shutdown()

    print(f"{COLOR_FRONTEND}[Frontend] Frontend server started successfully.{COLOR_RESET}")
    os.chdir(original_dir)
    return process

def prepare_backend_database(force=False):
    # script_dir = os.path.dirname(os.path.abspath(__file__))
    # backend_dir = os.path.join(script_dir, "backend")
    # db_file = os.path.join(backend_dir, "provenance.db")
    script_dir = os.path.dirname(os.path.abspath(__file__))
    project_root = os.path.abspath(os.path.join(script_dir, ".."))
    # backend_dir = os.path.join(script_dir, "backend")
    db_dir = os.path.join(os.getcwd(), ".curio")
    db_file = os.path.join(db_dir, "provenance.db")

    if not os.path.exists(db_file) or force:
        if not os.path.exists(db_dir):
            os.makedirs(db_dir)

        print(f"{COLOR_BACKEND}[Backend] Preparing backend database...{COLOR_RESET}")
        print(f"{COLOR_BACKEND}[Backend] Using database path: {db_file}{COLOR_RESET}")
        try:
            env = os.environ.copy()
            env = {**os.environ, "FLASK_APP": "utk_curio.backend.app:create_app", "PYTHONPATH": project_root + os.pathsep + env.get("PYTHONPATH", "")}
            subprocess.run(["python", "backend/create_provenance_db.py", os.path.abspath(db_file)], cwd=script_dir, check=True, env=env)

            subprocess.run(["flask", "db", "upgrade", "--directory", "utk_curio/backend/migrations"], check=True, cwd=project_root, env=env)
            subprocess.run(["flask", "db", "migrate", "-m", "Migration", "--directory", "utk_curio/backend/migrations"], check=True, cwd=project_root, env=env)
            print(f"{COLOR_BACKEND}[Backend] Database initialized successfully.{COLOR_RESET}")
        except Exception as e:
            print(f"{COLOR_BACKEND}[Backend] Failed to initialize the database: {e}{COLOR_RESET}")
            clean_shutdown()
    else:
        print(f"{COLOR_BACKEND}[Backend] Database already exists. Skipping initialization.{COLOR_RESET}")
    

def start_backend(host, port, force_db_init=False, no_server=False):
    print(f"Starting backend on {host}:{port}...")

    prepare_backend_database(force=force_db_init)

    # If we're only initializing the database, skip starting the server
    if no_server:
        print(f"{COLOR_BACKEND}[Backend] Database initialization completed with --force-db-init, server not started.{COLOR_RESET}")
        return None

    script_dir = os.path.dirname(os.path.abspath(__file__))
    project_root = os.path.abspath(os.path.join(script_dir, ".."))
    # backend_server = os.path.join(script_dir, "backend", "server.py")
    env = os.environ.copy()
    env = {**os.environ, "PYTHONPATH": project_root + os.pathsep + env.get("PYTHONPATH", "")}

    process = subprocess.Popen(
        ["python", "-u", "-m", "backend.server"],
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        text=True,
        cwd=script_dir,
        env=env
    )
    threading.Thread(target=stream_output, args=(process, "Backend", COLOR_BACKEND), daemon=True).start()
    return process


def start_sandbox(host, port):
    print(f"Starting sandbox on {host}:{port}...")

    script_dir = os.path.dirname(os.path.abspath(__file__))
    project_root = os.path.abspath(os.path.join(script_dir, ".."))
    # sandbox_server = os.path.join(script_dir, "sandbox", "server.py")
    env = os.environ.copy()
    env = {**os.environ, "PYTHONPATH": project_root + os.pathsep + env.get("PYTHONPATH", "")}

    process = subprocess.Popen(
        ["python", "-u", "-m", "sandbox.server"],
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        text=True,
        cwd=script_dir,
        env=env
    )
    threading.Thread(target=stream_output, args=(process, "Sandbox", COLOR_SANDBOX), daemon=True).start()
    return process

def signal_handler(sig, frame):
    """Handle SIGINT and SIGTERM for clean shutdown."""
    print("\nReceived shutdown signal (SIGINT or SIGTERM). Cleaning up...")
    clean_shutdown()

def clean_shutdown():
    print("\nShutting down all servers...")
    shutdown_flag.set()  # Signal threads to stop

    for process in processes:
        print(f"Terminating {process.args[0]}...")
        process.terminate()
        try:
            process.wait(timeout=5)
        except subprocess.TimeoutExpired:
            print(f"Force killing {process.args[0]}...")
            process.kill()

    print("All servers have been shut down.")
    # sys.stdout.flush()
    # sys.stderr.flush()
    sys.exit(0)  # Clean exit status (0)

def get_command_prefix():
    """Detects if the script is being run with 'python curio.py' or 'curio'."""
    if len(sys.argv) > 0:
        command = sys.argv[0]
        if command.endswith("curio.py"):
            return "python curio.py"
        elif command.endswith("curio"):
            return "curio"
    return "python curio.py"

def ensure_utk_installed():
    try:
        import utk
    except ImportError:
        print("[Main] Installing bundled utk tarball...")
        import subprocess, sys, os
        tarball_path = os.path.join(os.path.dirname(__file__), "sandbox", "utk-0.8.9.tar.gz")
        subprocess.check_call([sys.executable, "-m", "pip", "install", tarball_path])
        print("[Main] Installed utk successfully.")

def main():

    global processes

    # Capture Docker shutdown (SIGINT and SIGTERM)
    signal.signal(signal.SIGINT, signal_handler)
    signal.signal(signal.SIGTERM, signal_handler)

    command_prefix = get_command_prefix()

    parser = argparse.ArgumentParser(
        description="Curio's multi-server management tool.",
        formatter_class=argparse.RawTextHelpFormatter,
        epilog=f"""
    Examples:
        {command_prefix} start                       # Start all servers (backend, sandbox, frontend)
        {command_prefix} start backend               # Start only the backend (localhost:5002)
        {command_prefix} start sandbox               # Start only the sandbox (localhost:2000)
        {command_prefix} --force-rebuild             # Re-build the frontend (if dev mode)
        {command_prefix} --force-db-init             # Re-initialize the backend database (if dev mode)
    """
    )
    
    parser.add_argument(
        "command", nargs="?", choices=["start"], help="Command to execute (start)"
    )
    parser.add_argument(
        "server", nargs="?", default="all", choices=["all", "frontend", "backend", "sandbox"],
        help="Script to manage Curio's servers (all, frontend, backend, sandbox)"
    )
    parser.add_argument(
        "--backend-host", default="localhost", help="Host for the backend server (default: localhost)"
    )
    parser.add_argument(
        "--backend-port", default="5002", help="Port for the backend server (default: 5002)"
    )
    parser.add_argument(
        "--sandbox-host", default="localhost", help="Host for the sandbox server (default: localhost)"
    )
    parser.add_argument(
        "--sandbox-port", default="2000", help="Port for the sandbox server (default: 2000)"
    )
    if os.getenv("CURIO_DEV") == "1":
        parser.add_argument(
            "--force-rebuild", action="store_true",
            help="Force rebuild of the frontend"
        )
        parser.add_argument(
            "--force-db-init", action="store_true",
            help="Force re-initialization of the backend database"
        )

    # Display help if no arguments are given
    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit(0)

    args = parser.parse_args()

    set_environment_variables(
        backend_host=args.backend_host,
        backend_port=args.backend_port,
        sandbox_host=args.sandbox_host,
        sandbox_port=args.sandbox_port
    )

    # if os.getenv("CURIO_DEV") != "1":
        # if args.force_rebuild or args.force_db_init:
            # print("Error: --force-rebuild and --force-db-init are not available when running Curio from pip. If you really need it, refer to the documentation to run Curio from curio.py.")
            # sys.exit(1)
    if os.getenv("CURIO_DEV") == "1":
        # Handle standalone rebuild or db init without starting servers
        if not args.command:
            if args.force_rebuild:
                print("Rebuilding frontend...")
                start_frontend(force_rebuild=True, no_server=True)
            if args.force_db_init:
                print("Re-initializing backend database...")
                start_backend(args.backend_host, args.backend_port, force_db_init=True, no_server=True)
            sys.exit(0)
    else:
        args.force_rebuild = False
        args.force_db_init = False

    if args.command == "start":
        if args.server == "all":
            ensure_utk_installed()
            print("Starting all servers (backend, sandbox, frontend)...")
            processes = [
                start_backend(args.backend_host, args.backend_port, force_db_init=args.force_db_init),
                start_sandbox(args.sandbox_host, args.sandbox_port),
                start_frontend(force_rebuild=args.force_rebuild)
            ]
        else:
            if args.server == "backend":
                processes.append(start_backend(args.backend_host, args.backend_port))
            elif args.server == "sandbox":
                ensure_utk_installed()
                processes.append(start_sandbox(args.sandbox_host, args.sandbox_port))
            elif args.server == "frontend":
                processes.append(start_frontend(force_rebuild=args.force_rebuild))

        # Monitor the threads (logging simulation)
        logging_thread = threading.Thread(target=logger, daemon=True)
        logging_thread.start()

        try:
            while not shutdown_flag.is_set():
                time.sleep(1)
        except KeyboardInterrupt:
            clean_shutdown(processes)
    else:
        parser.print_help()

if __name__ == "__main__":
    main()


import json
import data

from osm import *
from urban_component import *
from load_physical import *
from load_thematic import *
from load_utk import *

import warnings
from shapely.errors import ShapelyDeprecationWarning
warnings.filterwarnings("ignore", category=ShapelyDeprecationWarning) 

def remove_elements(filepath, ids):
    file = open(filepath, mode='r')
    file_content = json.loads(file.read())

    new_data_array = []

    for index,data in enumerate(file_content['data']):
        if(index not in ids):
            new_data_array.append(data)

    file_content['data'] = new_data_array

    with open(filepath, "w") as outfile:
        json.dump(file_content, outfile)

    file.close()





"""Create UserSession model

Revision ID: f5c32a160bb4
Revises: 8273cbb189e4
Create Date: 2024-02-14 18:45:38.372036

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'f5c32a160bb4'
down_revision = '8273cbb189e4'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('user_session',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('user_id', sa.Integer(), nullable=False),
    sa.Column('token', sa.String(length=36), nullable=False),
    sa.Column('active', sa.Boolean(), nullable=False),
    sa.ForeignKeyConstraint(['user_id'], ['user.id'], ),
    sa.PrimaryKeyConstraint('id'),
    sa.UniqueConstraint('token')
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('user_session')
    # ### end Alembic commands ###
